from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    comps2unstg = _get_comps_to_unstage(args.comp_name, allcomps)
    for comp in comps2unstg:
        git = GitRepository(comp.remote, comp.local)
        staged_files = git.get_staged_files()
        for myfile in staged_files:
            git.unstage_file(myfile)
            print("- {}: {}".format(comp.name, myfile))


def _get_comps_to_unstage(specified_comps, allcomps):
    comps_to_unstage = allcomps
    if specified_comps:
        verify.valid_components(specified_comps, allcomps)
        comps_to_unstage = [x for x in allcomps if x.name in specified_comps]
    return comps_to_unstage



import os

from mepo.registry import Registry

TEST_DIR = os.path.dirname(os.path.realpath(__file__))


def get_ecbuild_details():
    return {
        "local": "./@cmake/@ecbuild",
        "remote": "../ecbuild.git",
        "tag": "geos/v1.2.0",
    }


def test_registry():
    registry = os.path.join(TEST_DIR, "input", "components.yaml")
    a = Registry(registry).read_file()
    assert a["ecbuild"] == get_ecbuild_details()



import os

from mepo.component import stylize_local_path
from mepo.component import MepoComponent
from mepo.registry import Registry
from mepo.utilities.version import MepoVersion

from _pytest.assertion import truncate

truncate.DEFAULT_MAX_LINES = 9999
truncate.DEFAULT_MAX_CHARS = 9999

TEST_DIR = os.path.dirname(os.path.realpath(__file__))


def get_registry():
    registry = os.path.join(TEST_DIR, "input", "components.yaml")
    return Registry(registry).read_file()


def get_fvdycore_component():
    return MepoComponent(
        name="fvdycore",
        local="./src/Components/@FVdycoreCubed_GridComp/@fvdycore",
        remote="https://github.com/GEOS-ESM/GFDL_atmos_cubed_sphere.git",
        version=MepoVersion(name="geos/v1.3.0", type="t", detached=True),
        sparse=None,
        develop="geos/develop",
        recurse_submodules=None,
        fixture=False,
        ignore_submodules=None,
    )


def get_fvdycore_serialized():
    return {
        "name": "fvdycore",
        "local": "./src/Components/@FVdycoreCubed_GridComp/@fvdycore",
        "remote": "https://github.com/GEOS-ESM/GFDL_atmos_cubed_sphere.git",
        "version": ["geos/v1.3.0", "t", True],
        "sparse": None,
        "develop": "geos/develop",
        "recurse_submodules": None,
        "fixture": False,
        "ignore_submodules": None,
    }


def test_stylize_local_path():
    local_path = "./src/Shared/@GMAO_Shared/@GEOS_Util"
    output = stylize_local_path(local_path, None)
    assert output == local_path
    output = stylize_local_path(local_path, "prefix")
    assert output == local_path
    output = stylize_local_path(local_path, "naked")
    assert output == "./src/Shared/@GMAO_Shared/GEOS_Util"
    output = stylize_local_path(local_path, "postfix")
    assert output == "./src/Shared/@GMAO_Shared/GEOS_Util@"


def test_MepoComponent():
    registry = get_registry()
    complist = list()
    for name, comp in registry.items():
        if name == "fvdycore":
            fvdycore = MepoComponent().registry_to_component(name, comp, None)
    assert fvdycore == get_fvdycore_component()
    fvdycore_serialized = fvdycore.serialize()
    remote = fvdycore_serialized["remote"]
    fvdycore_serialized["remote"] = remote.replace(
        "git@github.com:", "https://github.com/"
    )
    assert fvdycore_serialized == get_fvdycore_serialized()



from ..state import MepoState
from ..component import MepoVersion
from ..git import GitRepository
from ..utilities import colors


def run(args):
    allcomps = MepoState.read_state()
    detached_comps = []
    for comp in allcomps:
        git = GitRepository(comp.remote, comp.local)
        name, tYpe, detached = MepoVersion(*git.get_version())
        if detached:
            detached_comps.append(comp.name)
        else:
            print(
                "Pulling branch %s in %s "
                % (
                    colors.YELLOW + name + colors.RESET,
                    colors.RESET + comp.name + colors.RESET,
                )
            )
            output = git.pull()
            if not args.quiet:
                print(output)
    if len(detached_comps) > 0:
        print(
            "The following repos were not pulled (detached HEAD): %s"
            % (", ".join(map(str, detached_comps)))
        )



class MepoBranchArgParser:

    def __init__(self, branch):
        self.branch_subparsers = branch.add_subparsers()
        self.branch_subparsers.title = "mepo branch sub-commands"
        self.branch_subparsers.dest = "mepo_branch_cmd"
        self.branch_subparsers.required = True
        self.__list()
        self.__create()
        self.__delete()

    def __list(self):
        brlist = self.branch_subparsers.add_parser(
            "list",
            description="List local branches. If no component is specified, runs over all components",
        )
        brlist.add_argument(
            "-a", "--all", action="store_true", help="list all (local+remote) branches"
        )
        brlist.add_argument(
            "--nocolor", action="store_true", help="do not display color"
        )
        brlist.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="component to list branches in",
        )

    def __create(self):
        create = self.branch_subparsers.add_parser(
            "create", description="Create branch <branch-name> in component <comp-name>"
        )
        create.add_argument("branch_name", metavar="branch-name", help="name of branch")
        create.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="component to create branches in",
        )

    def __delete(self):
        delete = self.branch_subparsers.add_parser(
            "delete", description="Delete branch <branch-name> in component <comp-name>"
        )
        delete.add_argument("branch_name", metavar="branch-name", help="name of branch")
        delete.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="component to delete branches in",
        )
        delete.add_argument(
            "--force",
            action="store_true",
            help="delete branch even if it has not been fully merged",
        )



import os
import shlex

from dataclasses import dataclass
from urllib.parse import urljoin

from .git import get_current_remote_url
from .utilities import shellcmd
from .utilities.version import MepoVersion

# This will be used to store the "last nodes" from each subrepo
last_node_list = []


@dataclass(eq=True)
class MepoComponent(object):

    __slots__ = [
        "name",
        "local",
        "remote",
        "version",
        "sparse",
        "develop",
        "recurse_submodules",
        "fixture",
        "ignore_submodules",
    ]

    def __init__(
        self,
        name=None,
        local=None,
        remote=None,
        version=None,
        sparse=None,
        develop=None,
        recurse_submodules=None,
        fixture=None,
        ignore_submodules=None,
    ):
        self.name = name
        self.local = local
        self.remote = remote
        self.version = version
        self.sparse = sparse
        self.develop = develop
        self.recurse_submodules = recurse_submodules
        self.fixture = fixture
        self.ignore_submodules = ignore_submodules

    def __repr__(self):
        # Older mepo clones will not have ignore_submodules in comp, so
        # we need to handle this gracefully
        try:
            _ignore_submodules = self.ignore_submodules
        except AttributeError:
            _ignore_submodules = None

        return (
            f"{self.name} -\n"
            f"  local: {self.local}\n"
            f"  remote: {self.remote}\n"
            f"  version: {self.version}\n"
            f"  sparse: {self.sparse}\n"
            f"  develop: {self.develop}\n"
            f"  recurse_submodules: {self.recurse_submodules}\n"
            f"  fixture: {self.fixture}\n"
            f"  ignore_submodules: {_ignore_submodules}"
        )

    def __set_original_version(self, comp_details):
        if self.fixture:
            cmd_if_branch = "git symbolic-ref HEAD"
            # Have to use 'if not' since 0 is a good status
            if not shellcmd.run(cmd_if_branch.split(), status=True):
                output = shellcmd.run(cmd_if_branch.split(), output=True).rstrip()
                ver_name = output.replace("refs/heads/", "")
                ver_type = "b"
                is_detached = False
            else:
                # On some CI systems, git is handled oddly. As such, sometimes
                # tags aren't found due to shallow clones
                cmd_for_tag = "git describe --tags"
                # Have to use 'if not' since 0 is a good status
                if not shellcmd.run(cmd_for_tag.split(), status=True):
                    ver_name = shellcmd.run(cmd_for_tag.split(), output=True).rstrip()
                    ver_type = "t"
                    is_detached = True
                else:
                    # Per internet, describe always should always work, though mepo
                    # will return weirdness (a grafted branch, probably a hash)
                    cmd_for_always = "git describe --always"
                    ver_name = shellcmd.run(
                        cmd_for_always.split(), output=True
                    ).rstrip()
                    ver_type = "h"
                    is_detached = True
        else:
            if comp_details.get("branch", None):
                # SPECIAL HANDLING of 'detached head' branches
                ver_name = "origin/" + comp_details["branch"]
                ver_type = "b"
                # we always detach branches from components.yaml
                is_detached = True
            elif comp_details.get("hash", None):
                # Hashes don't have to exist
                ver_name = comp_details["hash"]
                ver_type = "h"
                is_detached = True
            else:
                ver_name = comp_details["tag"]  # 'tag' key has to exist
                ver_type = "t"
                is_detached = True
        self.version = MepoVersion(ver_name, ver_type, is_detached)

    def registry_to_component(self, comp_name, comp_details, comp_style):
        self.name = comp_name
        self.fixture = comp_details.get("fixture", False)
        # local/remote
        if self.fixture:
            self.local = "."
            self.remote = get_current_remote_url()
        else:
            self.local = stylize_local_path(comp_details["local"], comp_style)
            self.remote = comp_details["remote"]
            if self.remote.startswith("../"):
                self.remote = urljoin(get_current_remote_url() + "/", self.remote)
        # Optionals (None, if missing)
        self.sparse = comp_details.get("sparse", None)
        self.develop = comp_details.get("develop", None)
        self.recurse_submodules = comp_details.get("recurse_submodules", None)
        self.ignore_submodules = comp_details.get("ignore_submodules", None)
        # version
        self.__set_original_version(comp_details)

        return self

    def to_registry_format(self):
        details = dict()
        # Fixtures are allowed exactly two entries
        if self.fixture:
            details["fixture"] = self.fixture
            if self.develop:
                details["develop"] = self.develop
        else:
            details["local"] = self.local
            details["remote"] = self.remote
            if self.version.type == "t":
                details["tag"] = self.version.name
            elif self.version.type == "h":
                details["hash"] = self.version.name
            else:  # if not tag or hash, version has to be a branch
                if (
                    self.version.detached
                ):  # SPECIAL HANDLING of 'detached head' branches
                    details["branch"] = self.version.name.replace("origin/", "")
                else:
                    details["branch"] = self.version.name
            if self.sparse:
                details["sparse"] = self.sparse
            if self.develop:
                details["develop"] = self.develop
            if self.recurse_submodules:
                details["recurse_submodules"] = self.recurse_submodules
            if self.ignore_submodules:
                details["ignore_submodules"] = self.ignore_submodules
        return {self.name: details}

    def deserialize(self, d):
        for k in self.__slots__:
            v = d[k]
            if k == "version":
                # list -> namedtuple
                v = MepoVersion(*v)  # * for arg unpacking
            setattr(self, k, v)
        return self

    def serialize(self):
        d = {}
        for k in self.__slots__:
            v = getattr(self, k)
            if k == "version":
                # namedtuple -> list
                v = list(v)
            d.update({k: v})
        return d


def stylize_local_path(local_path, style):
    repo_flag = "@"  # Assumed flag for repos
    local_list = splitall(local_path)
    last_node = local_list[-1]
    last_node_list.append(last_node)  # maintain a list of last nodes
    # Decorate ALL last nodes since we can have nested repos
    for item in last_node_list:
        try:
            index = local_list.index(item)
            local_list[index] = decorate_node(item, repo_flag, style)
        except ValueError:
            pass
    return os.path.join(*local_list)


def decorate_node(item, flag, style):
    if not style:
        return item  # use what's in the registry file
    item = item.replace(flag, "")  # remove existing flag
    if style == "naked":
        output = item
    elif style == "prefix":
        output = flag + item
    elif style == "postfix":
        output = item + flag
    else:
        raise Exception(f"Invalid style: {style}")
    return output


# From https://learning.oreilly.com/library/view/python-cookbook/0596001673/ch04s16.html
def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts



"""contextlib.chdir was introduced in Python 3.11"""

import os
from contextlib import contextmanager


@contextmanager
def chdir(path):
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)



from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    verify.valid_components(args.comp_name, allcomps)
    comps2pushst = [x for x in allcomps if x.name in args.comp_name]
    for comp in comps2pushst:
        git = GitRepository(comp.remote, comp.local)
        git.push_stash(args.message)
        # print('+ {}'.format(comp.name))



from .config_get import run as config_get_run
from .config_set import run as config_set_run
from .config_delete import run as config_delete_run
from .config_print import run as config_print_run


def run(args):
    d = {
        "get": config_get_run,
        "set": config_set_run,
        "delete": config_delete_run,
        "print": config_print_run,
    }
    d[args.mepo_config_cmd](args)



from ..utilities import mepoconfig


def run(args):
    section, option = mepoconfig.split_entry(args.entry)
    if not mepoconfig.has_section(section):
        raise Exception(f"Section [{section}] does not exist in .mepoconfig")
    if not mepoconfig.has_option(section, option):
        raise Exception(
            f"Option [{option}] does not exist in section [{section}] in .mepoconfig"
        )
    value = mepoconfig.get(section, option)
    print(
        f"""
    [{section}]
    {option} = {value}
    """
    )



import multiprocessing as mp

from ..state import MepoState
from ..git import GitRepository
from ..utilities.version import version_to_string
from ..utilities import colors


def run(args):
    print("Checking status...", flush=True)
    allcomps = MepoState.read_state()
    if args.serial:
        result = []
        for comp in allcomps:
            result.append(check_component_status(comp))
    else:
        with mp.Pool() as pool:
            result = pool.map(check_component_status, allcomps)
    restore_state(allcomps, result)


def check_component_status(comp):
    git = GitRepository(comp.remote, comp.local)
    curr_ver = version_to_string(git.get_version(), git)
    return (curr_ver, git.check_status())


def restore_state(allcomps, result):
    for index, comp in enumerate(allcomps):
        git = GitRepository(comp.remote, comp.local)
        current_version = result[index][0].split(" ")[1]
        orig_version = comp.version.name
        if current_version != orig_version:
            print(
                colors.YELLOW
                + "Restoring "
                + colors.RESET
                + "{} to {} from {}.".format(
                    comp.name,
                    colors.GREEN + orig_version + colors.RESET,
                    colors.RED + current_version + colors.RESET,
                )
            )
            git.checkout(comp.version.name)



import configparser
import os
import sys

config_file = os.path.expanduser("~/.mepoconfig")
config = configparser.ConfigParser()
config.read(config_file)


def split_entry(entry):
    entry_list = entry.split(".")
    if len(entry_list) != 2:
        raise Exception(
            f'Invalid entry [{entry}]. Must be of form section.option, e.g., "alias.st"'
        )
    section = entry_list[0]
    option = entry_list[1]
    return section, option


def write():
    with open(config_file, "w") as fp:
        config.write(fp)


def print_sections():
    print(config.sections())


def print_options(section):
    print(config.options(section))


def print():
    config.write(sys.stdout)


def has_section(section):
    return config.has_section(section)


def has_option(section, option):
    return config.has_option(section, option)


def get(section, option):
    return config[section][option]


def remove_option(section, option):
    config.remove_option(section, option)
    if not config[section]:
        config.remove_section(section)


def set(section, option, value):
    if not has_section(section):
        config[section] = {}
    config[section][option] = value


def get_command_alias(command):
    output = []
    if has_section("alias"):
        for key, value in config.items("alias"):
            if value == command:
                output.append(key)
    return output


def get_alias_command(alias):
    command = alias
    if has_section("alias"):
        for key, value in config.items("alias"):
            if key == alias:
                command = value
    return command



import os
import io
import shutil
import contextlib
from types import SimpleNamespace

try:
    from contextlib import chdir as contextlib_chdir
except ImportError:
    from mepo.utilities.chdir import chdir as contextlib_chdir

import mepo.command.clone as mepo_clone
import mepo.command.status as mepo_status


FIXTURE_NAME = "GEOSfvdycore-mepo-testing"
FIXTURE_URL = f"https://github.com/pchakraborty/{FIXTURE_NAME}.git"
TEST_DIR = os.path.dirname(os.path.realpath(__file__))


def get_mepo_status():
    args = SimpleNamespace(
        ignore_permissions=False,
        nocolor=True,
        hashes=False,
        serial=True,
    )
    with contextlib.redirect_stdout(io.StringIO()) as output:
        mepo_status.run(args)
    return output.getvalue()


def test_mepo_clone_url():
    if os.path.isdir(FIXTURE_NAME):
        shutil.rmtree(FIXTURE_NAME)
    args = SimpleNamespace(
        style="prefix",
        registry=None,
        url=FIXTURE_URL,
        branch=None,
        directory=None,
        partial="blobless",
        allrepos=False,
    )
    mepo_clone.run(args)
    saved_output = """Checking status...
GEOSfvdycore           | (b) main
env                    | (t) v4.29.1 (DH)
cmake                  | (t) v3.55.0 (DH)
ecbuild                | (t) geos/v1.4.0 (DH)
GMAO_Shared            | (t) v1.9.9 (DH)
GEOS_Util              | (t) v2.1.3 (DH)
FMS                    | (t) geos/2019.01.02+noaff.10 (DH)
FVdycoreCubed_GridComp | (t) v2.12.0 (DH)
fvdycore               | (t) geos/v2.9.1 (DH)
mom6                   | (t) geos/v3.3 (DH)
"""
    with contextlib_chdir(FIXTURE_NAME):
        status_output = get_mepo_status()
    assert status_output == saved_output
    shutil.rmtree(FIXTURE_NAME)


def test_mepo_clone_url_branch_directory():
    DIRECTORY = "clone_url_branch_directory"
    if os.path.isdir(DIRECTORY):
        shutil.rmtree(DIRECTORY)
    args = SimpleNamespace(
        style="postfix",
        registry=None,
        url=FIXTURE_URL,
        branch="fvdycore-develop",
        directory=DIRECTORY,
        partial="blobless",
        allrepos=False,
    )
    mepo_clone.run(args)
    saved_output = """Checking status...
GEOSfvdycore           | (b) fvdycore-develop
env                    | (t) v4.29.1 (DH)
cmake                  | (t) v3.55.0 (DH)
ecbuild                | (t) geos/v1.4.0 (DH)
GMAO_Shared            | (t) v1.9.9 (DH)
GEOS_Util              | (t) v2.1.3 (DH)
FMS                    | (t) geos/2019.01.02+noaff.10 (DH)
FVdycoreCubed_GridComp | (b) develop
fvdycore               | (b) geos/develop
"""
    with contextlib_chdir(DIRECTORY):
        status_output = get_mepo_status()
    assert status_output == saved_output
    shutil.rmtree(DIRECTORY)


def test_mepo_clone_url_branch_allrepos():
    if os.path.isdir(FIXTURE_NAME):
        shutil.rmtree(FIXTURE_NAME)
    args = SimpleNamespace(
        style="prefix",
        registry=None,
        url=FIXTURE_URL,
        branch="mepo-testing-do-not-delete",
        directory=None,
        partial="blobless",
        allrepos=True,
    )
    mepo_clone.run(args)
    saved_output = """Checking status...
GEOSfvdycore           | (b) mepo-testing-do-not-delete
env                    | (b) mepo-testing-do-not-delete
cmake                  | (b) mepo-testing-do-not-delete
ecbuild                | (b) mepo-testing-do-not-delete
GMAO_Shared            | (b) mepo-testing-do-not-delete
GEOS_Util              | (b) mepo-testing-do-not-delete
FMS                    | (b) mepo-testing-do-not-delete
FVdycoreCubed_GridComp | (b) mepo-testing-do-not-delete
fvdycore               | (b) mepo-testing-do-not-delete
"""
    with contextlib_chdir(FIXTURE_NAME):
        status_output = get_mepo_status()
    assert status_output == saved_output
    shutil.rmtree(FIXTURE_NAME)


def test_mepo_clone_url_external_registry():
    if os.path.isdir(FIXTURE_NAME):
        shutil.rmtree(FIXTURE_NAME)
    args = SimpleNamespace(
        style="prefix",
        registry=os.path.join(TEST_DIR, "input/external-components.yaml"),
        url=FIXTURE_URL,
        branch="mepo-testing-do-not-delete",
        directory=None,
        partial="blobless",
        allrepos=False,
    )
    mepo_clone.run(args)
    saved_output = """Checking status...
GEOSfvdycore           | (b) mepo-testing-do-not-delete
   | external-components.yaml: \x1b[31muntracked file\x1b[0m
env                    | (t) v4.29.1 (DH)
cmake                  | (t) v3.55.0 (DH)
ecbuild                | (t) geos/v1.4.0 (DH)
GMAO_Shared            | (b) mepo-testing-do-not-delete
GEOS_Util              | (t) v2.1.3 (DH)
FMS                    | (t) geos/2019.01.02+noaff.10 (DH)
FVdycoreCubed_GridComp | (b) develop
fvdycore               | (b) geos/develop
"""
    with contextlib_chdir(FIXTURE_NAME):
        status_output = get_mepo_status()
    assert status_output == saved_output
    shutil.rmtree(FIXTURE_NAME)



try:
    from colorama import Fore, Style

    RED = Fore.RED
    BLUE = Fore.BLUE
    CYAN = Fore.CYAN
    GREEN = Fore.GREEN
    YELLOW = Fore.YELLOW
    RESET = Style.RESET_ALL
except ImportError:
    RED = "\x1b[1;31m"
    BLUE = "\x1b[1;34m"
    CYAN = "\x1b[1;36m"
    GREEN = "\x1b[1;32m"
    YELLOW = "\x1b[1;33m"
    RESET = "\x1b[0;0m"



from ..state import MepoState
from ..git import GitRepository
from ..utilities import colors


def run(args):
    allcomps = MepoState.read_state()
    for comp in allcomps:
        git = GitRepository(comp.remote, comp.local)
        ref_name = args.ref_name
        status, ref_type = git.verify_branch_or_tag(ref_name)

        if status == 0:
            if args.dry_run:
                print(
                    "%s %s exists in %s"
                    % (
                        ref_type,
                        colors.YELLOW + ref_name + colors.RESET,
                        colors.RESET + comp.name + colors.RESET,
                    )
                )
            else:
                if not args.quiet:
                    print(
                        "Checking out %s %s in %s"
                        % (
                            ref_type.lower(),
                            colors.YELLOW + ref_name + colors.RESET,
                            colors.RESET + comp.name + colors.RESET,
                        )
                    )
                git.checkout(ref_name, args.detach)



import os

from ..state import MepoState
from ..component import MepoVersion
from ..git import GitRepository
from ..registry import Registry
from ..utilities.version import sanitize_version_string


def run(args):
    allcomps = MepoState.read_state()
    for comp in allcomps:
        _update_comp(comp)

    MepoState.write_state(allcomps)

    complist = dict()
    relpath_start = MepoState.get_root_dir()
    for comp in allcomps:
        complist.update(comp.to_registry_format())
    registry_root_dir = os.path.join(relpath_start, args.registry)
    Registry(registry_root_dir).write_yaml(complist)
    print(f"Components written to '{registry_root_dir}'")


def _update_comp(comp):
    git = GitRepository(comp.remote, comp.local)
    orig_ver = comp.version
    curr_ver = MepoVersion(*git.get_version())

    orig_ver_is_tag_or_hash = orig_ver.type == "t" or orig_ver.type == "h"
    curr_ver_is_tag_or_hash = curr_ver.type == "t" or curr_ver.type == "h"

    if orig_ver_is_tag_or_hash and curr_ver_is_tag_or_hash:
        # This command is to try and work with git tag oddities
        curr_ver_to_use = sanitize_version_string(orig_ver.name, curr_ver.name, git)
        if curr_ver_to_use == orig_ver.name:
            comp.version = orig_ver
        else:
            _verify_local_and_remote_commit_ids_match(
                git, curr_ver_to_use, comp.name, curr_ver.type
            )
            comp.version = curr_ver
    else:
        if _version_has_changed(curr_ver, orig_ver, comp.name):
            _verify_local_and_remote_commit_ids_match(
                git, curr_ver.name, comp.name, curr_ver.type
            )
            comp.version = curr_ver


def _version_has_changed(curr_ver, orig_ver, name):
    result = False
    if curr_ver != orig_ver:
        if curr_ver.type == "b":
            assert (
                curr_ver.detached is False
            ), f"You cannot save a detached branch, have you committed your code in {name}?\n {curr_ver}"
            result = True
        elif curr_ver.type == "t":
            result = True
        elif curr_ver.type == "h":
            result = True
        else:
            raise Exception("This should not happen")
    return result


def _verify_local_and_remote_commit_ids_match(
    git, curr_ver_name, comp_name, curr_ver_type
):
    remote_id = git.get_remote_latest_commit_id(curr_ver_name, curr_ver_type)
    local_id = git.get_local_latest_commit_id()
    failmsg = (
        "{} (remote commit) != {} (local commit) for {}:{}. Did you try 'mepo push'?"
    )
    if remote_id != local_id:
        msg = failmsg.format(remote_id, local_id, comp_name, curr_ver_name)
        raise Exception(msg)



class MepoTagArgParser:

    def __init__(self, tag):
        self.tag = tag.add_subparsers()
        self.tag.title = "mepo tag sub-commands"
        self.tag.dest = "mepo_tag_cmd"
        self.tag.required = True
        self.__list()
        self.__create()
        self.__delete()
        self.__push()

    def __list(self):
        tglist = self.tag.add_parser(
            "list",
            description="List tags. If no component is specified, runs over all components",
        )
        tglist.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="Component to list tags in",
        )

    def __create(self):
        create = self.tag.add_parser(
            "create",
            description="Create tag <tag-name> in component <comp-name>. If no component is specified, runs over all components",
        )
        create.add_argument("tag_name", metavar="tag-name", help="Name of tag")
        create.add_argument(
            "-a", "--annotate", action="store_true", help="Make an annotated tag"
        )
        create.add_argument(
            "-m",
            "--message",
            type=str,
            metavar="message",
            default=None,
            help="Message for the tag",
        )
        create.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="Component to create tags in",
        )

    def __delete(self):
        delete = self.tag.add_parser(
            "delete",
            description="Delete tag <tag-name> in component <comp-name>. If no component is specified, runs over all components",
        )
        delete.add_argument("tag_name", metavar="tag-name", help="Name of tag")
        delete.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="Component to delete tags in",
        )

    def __push(self):
        push = self.tag.add_parser(
            "push",
            description="Push tag <tag-name> in component <comp-name>. If no component is specified, runs over all components",
        )
        push.add_argument("tag_name", metavar="tag-name", help="Name of tag")
        push.add_argument(
            "-f", "--force", action="store_true", help="Force push (be careful!)"
        )
        push.add_argument(
            "-d", "--delete", action="store_true", help="Delete (be careful!)"
        )
        push.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="Component to push tags in",
        )



from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    verify.valid_components(args.comp_name, allcomps)
    comps2delbr = [x for x in allcomps if x.name in args.comp_name]
    for comp in comps2delbr:
        git = GitRepository(comp.remote, comp.local)
        git.delete_branch(args.branch_name, args.force)
        print("- {}: {}".format(comp.name, args.branch_name))



import os
import shutil

from ..state import MepoState
from ..utilities.exceptions import NotInRootDirError

from .clone import run as mepo_clone_run

# This command will "reset" the mepo clone. This will delete all
# the subrepos, remove the mepo state directory, and then re-clone all the
# subrepos. This is useful if you want to start over with a fresh clone
# of the project.


def run(args):
    allcomps = MepoState.read_state()

    # Check to see that we are in the root directory of the project
    ## First get the root directory of the project
    rootdir = MepoState.get_root_dir()
    ## Then get the current directory
    curdir = os.getcwd()
    ## Then check that they are the same, if they are not, then throw a NotInRootDirError
    if rootdir != curdir:
        raise NotInRootDirError(
            "Error! As a safety precaution, you must be in the root directory of the project to reset"
        )

    # If we get this far, then we are in the root directory of the project

    # If a user has called this command without the force flag, we
    # will ask them to confirm that they want to reset the project
    if not args.force and not args.dry_run:
        print(
            "Are you sure you want to reset the project? If so, type 'yes' and press enter.",
            end=" ",
        )
        answer = input()
        if answer != "yes":
            print("Reset cancelled.")
            return

    # First, we need to delete all the subrepos
    # Loop over all the components in reverse (since we are deleting them)
    for comp in reversed(allcomps):
        # If the component is the Fixture, then skip it
        if comp.fixture:
            continue
        else:
            # Get the relative path to the component
            relpath = _get_relative_path(comp.local)
            print(f"Removing {relpath}", end="...")
            # Remove the component if not dry run
            if not args.dry_run:
                shutil.rmtree(relpath)
                print("done.")
            else:
                print(f"dry-run only. Not removing {relpath}")

    # Next, we need to remove the .mepo directory
    print("Removing mepo state", end="...")
    if not args.dry_run:
        shutil.rmtree(MepoState.get_dir())
        print("done.")
    else:
        print("dry-run only. Not removing mepo state")

    # If they pass in the --reclone flag, then we will re-clone all the subrepos
    if args.reclone:

        # mepo_clone requires args which is an Argparse Namespace object
        # We will create a new Namespace object with the correct arguments
        # for mepo_clone
        clone_args = type(
            "Namespace",
            (object,),
            {
                "url": None,
                "directory": None,
                "branch": None,
                "registry": None,
                "allrepos": False,
                "style": None,
            },
        )
        if not args.dry_run:
            print("Re-cloning all subrepos")
            mepo_clone_run(clone_args)
            print("Recloning done.")
        else:
            print("Dry-run only. Not re-cloning all subrepos")


def _get_relative_path(local_path):
    """
    Get the relative path when given a local path.

    local_path: The path to a subrepo as known by mepo (relative to the mepo state directory)
    """

    # This creates a full path on the disk from the root of mepo and the local_path
    full_local_path = os.path.join(MepoState.get_root_dir(), local_path)

    # We return the path relative to where we currently are
    return os.path.relpath(full_local_path, os.getcwd())



import os

import io
import shutil
import shlex
import unittest
import importlib
import contextlib
import subprocess as sp
from types import SimpleNamespace

import mepo.command.clone as mepo_clone
import mepo.command.list as mepo_list
import mepo.command.status as mepo_status
import mepo.command.compare as mepo_compare
import mepo.command.develop as mepo_develop
import mepo.command.checkout as mepo_checkout
import mepo.command.branch_list as mepo_branch_list
import mepo.command.branch_create as mepo_branch_create
import mepo.command.branch_delete as mepo_branch_delete
import mepo.command.tag_list as mepo_tag_list
import mepo.command.tag_create as mepo_tag_create
import mepo.command.tag_delete as mepo_tag_delete
import mepo.command.fetch as mepo_fetch
import mepo.command.pull as mepo_pull
import mepo.command.push as mepo_push
import mepo.command.diff as mepo_diff
import mepo.command.whereis as mepo_whereis
import mepo.command.reset as mepo_reset
from mepo.cmdline.parser import get_version as get_mepo_version

# Import commands with dash in the name
mepo_restore_state = importlib.import_module("mepo.command.restore-state")
mepo_checkout_if_exists = importlib.import_module("mepo.command.checkout-if-exists")
mepo_pull_all = importlib.import_module("mepo.command.pull-all")

TEST_DIR = os.path.dirname(os.path.realpath(__file__))


class TestMepoCommands(unittest.TestCase):

    @classmethod
    def __get_saved_output(cls, output_file):
        with open(os.path.join(cls.output_dir, output_file), "r") as fin:
            saved_output = fin.read()
        return saved_output

    @classmethod
    def __checkout_fixture(cls):
        remote = f"https://github.com/GEOS-ESM/{cls.fixture}.git"
        git_clone = "git clone --filter=blob:none "
        if cls.tag:
            git_clone += f"-b {cls.tag}"
        cmd = f"{git_clone} {remote} {cls.fixture_dir}"
        sp.run(shlex.split(cmd))

    @classmethod
    def __copy_config_file(cls):
        src = os.path.join(cls.input_dir, "components.yaml")
        dst = os.path.join(cls.fixture_dir)
        shutil.copy(src, dst)

    @classmethod
    def __mepo_clone(cls):
        # mepo clone
        args = SimpleNamespace(
            style="prefix",
            registry=None,
            url=None,
            allrepos=False,
            branch=None,
            directory=None,
            partial="blobless",
        )
        mepo_clone.run(args)
        print(flush=True)

    @classmethod
    def setUpClass(cls):
        cls.input_dir = os.path.join(TEST_DIR, "input")
        cls.output_dir = os.path.join(TEST_DIR, "output")
        cls.output_clone_status = cls.__get_saved_output("output_clone_status.txt")
        cls.fixture = "GEOSfvdycore"
        cls.tag = "v2.13.0"
        cls.tmpdir = os.path.join(TEST_DIR, "tmp")
        cls.fixture_dir = os.path.join(cls.tmpdir, cls.fixture)
        if os.path.isdir(cls.fixture_dir):
            shutil.rmtree(cls.fixture_dir)
        cls.__checkout_fixture()
        os.chdir(cls.fixture_dir)
        cls.__mepo_clone()

    def setUp(self):
        pass

    def __mepo_status(self, saved_output):
        """saved_output is either a string or a filename"""
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            ignore_permissions=False,
            nocolor=True,
            hashes=False,
            serial=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_status.run(args)
        try:  # assume saved_output is a file
            saved_output_s = self.__class__.__get_saved_output(saved_output)
        except FileNotFoundError:
            saved_output_s = saved_output
        self.assertEqual(output.getvalue(), saved_output_s)

    def __mepo_restore_state(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(serial=False)
        with contextlib.redirect_stdout(io.StringIO()) as _:
            mepo_restore_state.run(args)
        self.__mepo_status(self.__class__.output_clone_status)

    def test_list(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(one_per_line=False)
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_list.run(args)
        saved_output = self.__class__.__get_saved_output("output_list.txt")
        self.assertEqual(output.getvalue(), saved_output)

    def test_develop(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            comp_name=["env", "cmake", "fvdycore"],
            quiet=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as _:
            mepo_develop.run(args)
        self.__mepo_status("output_develop_status.txt")
        # Clean up
        self.__mepo_restore_state()

    def test_checkout_compare(self):
        os.chdir(self.__class__.fixture_dir)
        # Checkout "develop" branch of MAPL and env
        args = SimpleNamespace(
            branch_name="develop",
            comp_name=["MAPL"],
            b=False,
            quiet=False,
            detach=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_checkout.run(args)
        # Compare (default)
        args_cmp = SimpleNamespace(
            all=False,
            nocolor=True,
            wrap=True,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_compare.run(args_cmp)
        saved_output = self.__class__.__get_saved_output("output_compare.txt")
        self.assertEqual(output.getvalue(), saved_output)
        # Compare (All)
        args_cmp.all = True
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_compare.run(args_cmp)
        saved_output = self.__class__.__get_saved_output("output_compare_all.txt")
        self.assertEqual(output.getvalue(), saved_output)
        # Clean up
        self.__mepo_restore_state()

    def test_checkout_if_exists(self):
        # Fixture component
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            ref_name="not-expected-to-exist",
            quiet=True,
            detach=False,
            dry_run=False,
        )
        mepo_checkout_if_exists.run(args)
        # Since we do not expect this ref to exist, status should be that of clone
        self.__mepo_status(self.__class__.output_clone_status)

    def test_branch_list(self):
        os.chdir(self.__class__.fixture_dir)
        # Not expecting new branches in this component (fingers crossed)
        args = SimpleNamespace(
            comp_name=["ecbuild"],
            all=True,
            nocolor=True,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_branch_list.run(args)
        saved_output = self.__class__.__get_saved_output("output_branch_list.txt")
        self.assertEqual(output.getvalue(), saved_output)

    def test_branch_create_delete(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            comp_name=["ecbuild"],
            branch_name="the-best-branch-ever",
        )
        # Create branch
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_branch_create.run(args)
        saved_output = self.__class__.__get_saved_output("output_branch_create.txt")
        self.assertEqual(output.getvalue(), saved_output)
        # Delete the branch that was just created
        args.force = False
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_branch_delete.run(args)
        saved_output = self.__class__.__get_saved_output("output_branch_delete.txt")
        self.assertEqual(output.getvalue(), saved_output)

    def test_tag_list(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(comp_name=["env"])
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_tag_list.run(args)
        self.assertTrue("cuda11.7.0-gcc11.2.0nvptx-openmpi4.0.6" in output.getvalue())

    def test_tag_create_delete(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            comp_name=["FMS", "MAPL"],
            tag_name="new-awesome-tag",
            annotate=False,
            message=None,
        )
        # Create tag
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_tag_create.run(args)
        saved_output = self.__class__.__get_saved_output("output_tag_create.txt")
        self.assertEqual(output.getvalue(), saved_output)
        # Delete the tag that was just created
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_tag_delete.run(args)
        saved_output = self.__class__.__get_saved_output("output_tag_delete.txt")
        self.assertEqual(output.getvalue(), saved_output)

    def test_fetch(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            comp_name=["FVdycoreCubed_GridComp"],
            all=True,
            prune=True,
            tags=True,
            force=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_fetch.run(args)
        saved_output = "Fetching \x1b[33mFVdycoreCubed_GridComp\x1b[0m\n"
        self.assertEqual(output.getvalue(), saved_output)

    def test_pull(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            comp_name=["FVdycoreCubed_GridComp"],
            quiet=False,
        )
        err_msg = "FVdycoreCubed_GridComp has detached head! Cannot pull."
        with self.assertRaisesRegex(Exception, err_msg):
            mepo_pull.run(args)

    def test_pull_all(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            comp_name=["FVdycoreCubed_GridComp"],
            quiet=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_pull_all.run(args)
        saved_output = self.__class__.__get_saved_output("output_pull_all.txt")
        self.assertEqual(output.getvalue(), saved_output)

    def test_push(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            comp_name=["FVdycoreCubed_GridComp"],
            quiet=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            with self.assertRaises(sp.CalledProcessError):
                mepo_push.run(args)
        saved_output = self.__class__.__get_saved_output("output_push.txt")
        self.assertEqual(output.getvalue(), saved_output)

    def test_diff(self):
        os.chdir(self.__class__.fixture_dir)
        os.chdir("./src/Components/@FVdycoreCubed_GridComp")
        filename = "GEOS_FV3_Utilities.F90"
        # Add a line
        with open(filename, "w") as fout:
            fout.write(" ")
        args = SimpleNamespace(
            comp_name=["FVdycoreCubed_GridComp"],
            name_only=True,
            name_status=False,
            ignore_permissions=False,
            staged=False,
            ignore_space_change=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_diff.run(args)
        saved_output = self.__class__.__get_saved_output("output_diff.txt")
        # Ignore the last line of output (horizontal line
        # with length that of the width of the terminal)
        self.assertEqual(output.getvalue().split()[:-1], saved_output.split()[:-1])
        # Clean up
        sp.run(f"git checkout {filename}".split(), stderr=sp.DEVNULL)
        self.__mepo_status(self.__class__.output_clone_status)

    def test_whereis(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            comp_name=None,
            ignore_case=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_whereis.run(args)
        saved_output = self.__class__.__get_saved_output("output_whereis.txt")
        self.assertEqual(output.getvalue(), saved_output)

    def test_reset(self):
        os.chdir(self.__class__.fixture_dir)
        args = SimpleNamespace(
            force=True,
            reclone=False,
            dry_run=False,
        )
        with contextlib.redirect_stdout(io.StringIO()) as output:
            mepo_reset.run(args)
        saved_output = self.__class__.__get_saved_output("output_reset.txt")
        self.assertEqual(output.getvalue(), saved_output)
        # Clean up - reclone (suppress output)
        with contextlib.redirect_stdout(io.StringIO()) as output:
            self.__class__.__mepo_clone()

    def test_mepo_version(self):
        self.assertEqual(get_mepo_version(), "2.2.1")

    def tearDown(self):
        pass

    @classmethod
    def tearDownClass(cls):
        os.chdir(TEST_DIR)
        shutil.rmtree(cls.tmpdir)


if __name__ == "__main__":
    unittest.main()



from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository
from ..utilities import colors


def run(args):
    allcomps = MepoState.read_state()
    verify.valid_components(args.comp_name, allcomps)
    comps2dev = [x for x in allcomps if x.name in args.comp_name]
    for comp in comps2dev:
        git = GitRepository(comp.remote, comp.local)
        if comp.develop is None:
            raise Exception("'develop' branch not specified for {}".format(comp.name))
        if not args.quiet:
            print(
                "Checking out development branch %s in %s"
                % (
                    colors.YELLOW + comp.develop + colors.RESET,
                    colors.RESET + comp.name + colors.RESET,
                )
            )
        git.checkout(comp.develop)
        _ = git.pull()



#!/usr/bin/env python3

import os
import glob
from mdutils.mdutils import MdUtils
import subprocess as sp

preamble = """
mepo provides many different commands for working with a multi-repository fixture.
"""

# Assume this script is in mepo/doc. Then we need to get to the mepo/mepo.d/command directory
doc_dir_path = os.path.dirname(os.path.realpath(__file__))
# Then we need to get to the mepo/mepo.d/command directory. First the "main" dir
main_dir_path = os.path.dirname(doc_dir_path)
# Now add "src/mepo"
mepod_dir_path = os.path.join(main_dir_path, "src", "mepo")
# And then "command"
command_dir_path = os.path.join(mepod_dir_path, "command")

mepo_command_path = os.path.join(main_dir_path, "bin", "mepo")


def get_command_list(directory):
    # Get all commands
    all_commands_py = glob.glob(os.path.join(directory, "*.py"))
    all_commands = [os.path.basename(x).replace(".py", "") for x in all_commands_py]

    # Now let's find the commands that have subcommands
    ## First we get commands with underscore
    ## Then replace underscore with a space
    commands_with_underscore = [x for x in all_commands if "_" in x]
    commands_with_subcommands = [x.replace("_", " ") for x in commands_with_underscore]
    all_useful_commands = [x for x in all_commands if x not in commands_with_underscore]
    all_useful_commands += commands_with_subcommands

    return sorted(all_useful_commands)


def create_markdown_from_usage(command, mdFile):
    cmd = [mepo_command_path, command, "--help"]

    # Some commands have spaces, so we need to break it up again
    cmd = " ".join(cmd).split()

    result = sp.run(cmd, capture_output=True, universal_newlines=True)
    output = result.stdout

    output_list = output.split("\n")

    # Command summary
    summary = output_list[2]
    mdFile.write(summary)

    # Usage
    usage = output_list[0]
    usage = usage.replace("usage: ", "")
    mdFile.new_header(level=3, title="Usage")
    mdFile.insert_code(usage)

    positional_arguments = output.partition("positional arguments:\n")[2].partition(
        "\n\n"
    )[0]
    if positional_arguments:
        mdFile.new_header(level=3, title="Positional Arguments")
        mdFile.insert_code(positional_arguments)

    optional_arguments = output.partition("optional arguments:\n")[2].partition("\n\n")[
        0
    ]
    # Remove extra blank lines
    optional_arguments = os.linesep.join(
        [s for s in optional_arguments.splitlines() if s]
    )
    if optional_arguments:
        mdFile.new_header(level=3, title="Optional Arguments")
        mdFile.insert_code(optional_arguments)


if __name__ == "__main__":

    doc_file = "Mepo-Commands.md"
    mdFile = MdUtils(file_name=doc_file)

    mdFile.new_header(level=1, title="Overview")
    mdFile.new_paragraph(preamble)

    mdFile.new_header(level=1, title="Commands")
    command_list = get_command_list(command_dir_path)
    for command in command_list:
        mdFile.new_header(level=2, title=command)
        print(f"mepo command: {command}")
        create_markdown_from_usage(command, mdFile)

    mdFile.new_table_of_contents(table_title="Table of Contents", depth=2)
    mdFile.create_md_file()
    print(f"Generated {doc_file}.")



from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    verify.valid_components(args.comp_name, allcomps)
    comps2tagpush = _get_comps_to_list(args.comp_name, allcomps)
    for comp in comps2tagpush:
        git = GitRepository(comp.remote, comp.local)
        git.push_tag(args.tag_name, args.force, args.delete)
        if args.delete:
            print(f"Pushed deleted tag {args.tag_name} to {comp.name}")
        else:
            print(f"Pushed tag {args.tag_name} to {comp.name}")


def _get_comps_to_list(specified_comps, allcomps):
    comps_to_list = allcomps
    if specified_comps:
        verify.valid_components(specified_comps, allcomps)
        comps_to_list = [x for x in allcomps if x.name in specified_comps]
    return comps_to_list



def valid_components(specified_comp_names, allcomps, ignore_case=False):
    """
    Validate a component name

    Arguments:
        specified_comp_names: List of component names
        allcomps: List of all components known to mepo

    Keyword arguments:
        ignore_case: should the comparison ignore case
    """

    # Make a list of all the component names depending on ignore_case
    all_component_names = [
        x.name.casefold() if ignore_case else x.name for x in allcomps
    ]

    # Loop over all the components we want to verify...
    for component_name in specified_comp_names:

        # Create a name to compare with based on ignore_case
        component_to_find = component_name.casefold() if ignore_case else component_name

        # Validate the component
        _validate_component(component_to_find, all_component_names)


def _validate_component(component, all_components):
    """
    Function to raise exception on invalid component name

    Arguments:
        component: component to validate
        all_components: List of valid components
    """

    if component not in all_components:
        raise Exception("Unknown component name [{}]".format(component))



from ..state import MepoState


def run(args):
    _end = "\n" if args.one_per_line else " "
    allcomps = MepoState.read_state()
    for comp in allcomps[:-1]:
        print(comp.name, end=_end)
    print(allcomps[-1].name)



import yaml
import pathlib

from .utilities.exceptions import SuffixNotRecognizedError


# From https://github.com/yaml/pyyaml/issues/127#issuecomment-525800484
class AddBlankLinesDumper(yaml.SafeDumper):
    # HACK: insert blank lines between top-level objects
    # inspired by https://stackoverflow.com/a/44284819/3786245
    def write_line_break(self, data=None):
        super().write_line_break(data)

        if len(self.indents) == 1:
            super().write_line_break()


class Registry(object):

    __slots__ = ["__filename", "__filetype"]

    def __init__(self, filename):
        self.__filename = filename
        SUFFIX_LIST = [".yaml", ".json", ".cfg"]
        file_suffix = pathlib.Path(filename).suffix
        if file_suffix in SUFFIX_LIST:
            self.__filetype = file_suffix[1:]
        else:
            raise SuffixNotRecognizedError(
                "suffix {} not supported".format(file_suffix)
            )

    def __validate(self, d):
        git_tag_types = {"branch", "tag", "hash"}
        num_fixtures = 0
        for k, v in d.items():
            if "fixture" in v:
                # In case of a fixture, develop is the only additional key
                num_fixtures += 1
                assert list(v.keys()) == ["fixture", "develop"]
            else:
                # For non-fixture, one and only one of branch/tag/hash allowed
                xsection = git_tag_types.intersection(set(v.keys()))
                if len(xsection) != 1:
                    raise ValueError(f"{k} needs one and only one of {git_tag_types}")
        # Can have one and only one fixture
        assert num_fixtures == 1

    def read_file(self):
        """Call read_yaml, read_json etc. using dispatch pattern"""
        return getattr(self, "read_" + self.__filetype)()

    def read_yaml(self):
        """Read yaml registry and return a dict containing contents"""
        import yaml

        with open(self.__filename, "r") as fin:
            d = yaml.safe_load(fin)
        self.__validate(d)
        return d

    def read_json(self):
        """Read json registry and return a dict containing contents"""
        import json

        with open(self.__filename, "r") as fin:
            d = json.load(fin)
        self.__validate(d)
        return d

    def read_cfg(self):
        """Read python registry and return a dict containing contents"""
        raise NotImplementedError("Reading of cfg file has not yet been implemented")

    def write_yaml(self, d):
        """Dump dict d into a yaml file"""
        import yaml

        with open(self.__filename, "w") as fout:
            yaml.dump(d, fout, sort_keys=False, Dumper=AddBlankLinesDumper)



from .tag_list import run as tag_list_run
from .tag_create import run as tag_create_run
from .tag_delete import run as tag_delete_run
from .tag_push import run as tag_push_run


def run(args):
    d = {
        "list": tag_list_run,
        "create": tag_create_run,
        "delete": tag_delete_run,
        "push": tag_push_run,
    }
    d[args.mepo_tag_cmd](args)



from .stash_list import run as stash_list_run
from .stash_pop import run as stash_pop_run
from .stash_apply import run as stash_apply_run
from .stash_push import run as stash_push_run
from .stash_show import run as stash_show_run


def run(args):
    d = {
        "list": stash_list_run,
        "pop": stash_pop_run,
        "apply": stash_apply_run,
        "push": stash_push_run,
        "show": stash_show_run,
    }
    d[args.mepo_stash_cmd](args)



class StateDoesNotExistError(SystemExit):
    """Raised when the mepo state does not exist"""

    pass


class StateAlreadyInitializedError(SystemExit):
    """Raised when the mepo state has already been initialized"""

    pass


class RepoAlreadyClonedError(SystemExit):
    """Raised when the repository has already been cloned"""

    pass


class RegistryNotFoundError(FileNotFoundError):
    """Raised when the registry is not found"""

    pass


class SuffixNotRecognizedError(RuntimeError):
    """Raised when the registry suffix is not recognized"""

    pass


class NotInRootDirError(SystemExit):
    """Raised when a command is run not in the root directory"""

    pass



import os

from ..state import MepoState

from ..utilities import verify
from ..utilities.version import version_to_string
from ..utilities.version import sanitize_version_string
from ..git import GitRepository

VER_LEN = 30


def run(args):
    allcomps = MepoState.read_state()

    if any_differing_repos(allcomps):
        comps2diff = _get_comps_to_diff(args.comp_name, allcomps)

        for comp in comps2diff:
            git = GitRepository(comp.remote, comp.local)
            orig_ver = version_to_string(comp.version, git).split()[1]
            orig_type = comp.version.type
            changed_files = git.get_changed_files(
                untracked=True, orig_ver=orig_ver, comp_type=orig_type
            )

            # If there are changed files, print them with the local path
            if changed_files:
                for file in changed_files:
                    if args.full_path:
                        print(os.path.abspath(os.path.join(comp.local, file)))
                    else:
                        print(os.path.join(comp.local, file))


def _get_comps_to_diff(specified_comps, allcomps):
    comps_to_diff = allcomps
    if specified_comps:
        verify.valid_components(specified_comps, allcomps)
        comps_to_diff = [x for x in allcomps if x.name in specified_comps]
    return comps_to_diff


def any_differing_repos(allcomps):
    for comp in allcomps:
        git = GitRepository(comp.remote, comp.local)
        curr_ver = version_to_string(git.get_version(), git)
        orig_ver = version_to_string(comp.version, git)

        # This command is to try and work with git tag oddities
        curr_ver = sanitize_version_string(orig_ver, curr_ver, git)

        if curr_ver not in orig_ver:
            return True

    return False



import os

from ..state import MepoState
from ..utilities import verify


def run(args):
    allcomps = MepoState.read_state()
    if args.comp_name:  # single comp name is specified, print relpath
        if args.comp_name == "_root":
            # _root is a "hidden" allowed argument for whereis to return
            # the root dir of the project. Mainly used by mepo-cd
            print(MepoState.get_root_dir())
        else:
            # Verify that we passed in a good component name
            verify.valid_components(
                [args.comp_name], allcomps, ignore_case=args.ignore_case
            )

            # Create a name to look for according to ignore_case option
            component_to_find = (
                args.comp_name.casefold() if args.ignore_case else args.comp_name
            )

            # Loop over all the components that mepo knows about...
            for comp in allcomps:
                # Create a name to compare to based on ignore_case option
                component_in_allcomps = (
                    comp.name.casefold() if args.ignore_case else comp.name
                )
                # And if they match, print the relative path
                if component_in_allcomps == component_to_find:
                    print(_get_relative_path(comp.local))

    else:  # print relpaths of all comps
        max_namelen = len(max([x.name for x in allcomps], key=len))
        FMT = "{:<%s.%ss} | {:<s}" % (max_namelen, max_namelen)
        for comp in allcomps:
            print(FMT.format(comp.name, _get_relative_path(comp.local)))


def _get_relative_path(local_path):
    """
    Get the relative path when given a local path.

    local_path: The path to a subrepo as known by mepo (relative to the .mepo directory)
    """

    # This creates a full path on the disk from the root of mepo and the local_path
    full_local_path = os.path.join(MepoState.get_root_dir(), local_path)

    # We return the path relative to where we currently are
    return os.path.relpath(full_local_path, os.getcwd())



import os
import tempfile
import subprocess

from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository
from ..git import get_editor as get_git_editor


def run(args):
    allcomps = MepoState.read_state()
    comps2crttg = _get_comps_to_list(args.comp_name, allcomps)

    tf_file = None

    if args.annotate:
        create_annotated_tag = True
    elif args.message:
        create_annotated_tag = True
    else:
        create_annotated_tag = False

    if create_annotated_tag:
        # Pop up an editor if a message is not provided
        # Popping up an EDITOR is based on https://stackoverflow.com/a/39989442
        if not args.message:
            EDITOR = get_git_editor()
            initial_message = b""  # set up the file

            # Use delete=False to keep the file around as we send the file name to git commit -F
            tf = tempfile.NamedTemporaryFile(delete=False)
            tf_file = tf.name
            tf.write(initial_message)
            tf.flush()
            subprocess.call([EDITOR, tf.name])

    for comp in comps2crttg:
        git = GitRepository(comp.remote, comp.local)
        git.create_tag(args.tag_name, create_annotated_tag, args.message, tf_file)
        print("+ {}: {}".format(comp.name, args.tag_name))

    if create_annotated_tag:
        # Now close and by-hand delete the temp file
        if not args.message:
            tf.close()
            os.unlink(tf.name)


def _get_comps_to_list(specified_comps, allcomps):
    comps_to_list = allcomps
    if specified_comps:
        verify.valid_components(specified_comps, allcomps)
        comps_to_list = [x for x in allcomps if x.name in specified_comps]
    return comps_to_list



from ..state import MepoState
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    max_namelen = len(max([x.name for x in allcomps], key=len))
    FMT = "{:<%s.%ss} | {:<s}" % (max_namelen, max_namelen)
    for comp in allcomps:
        git = GitRepository(comp.remote, comp.local)
        output = git.list_stash().rstrip().split("\n")
        print(FMT.format(comp.name, output[0]))
        for line in output[1:]:
            print(FMT.format("", line))



"""Current state of mepo managed repositories"""

import time
import shlex
import multiprocessing as mp

from ..state import MepoState
from ..git import GitRepository
from ..utilities import colors
from ..utilities import shellcmd
from ..utilities.version import version_to_string
from ..utilities.version import sanitize_version_string

from .whereis import _get_relative_path


def run(args):
    """Entry point"""
    print("Checking status...", flush=True)
    allcomps = MepoState.read_state()
    # max_width = len(max([comp.name for comp in allcomps], key=len))
    max_width = max([len(comp.name) for comp in allcomps])
    if args.serial:
        for comp in allcomps:
            result = check_component_status(comp, args.ignore_permissions)
            print_component_status(comp, result, max_width, args.nocolor, args.hashes)
    else:
        with mp.Pool() as pool:
            result = pool.starmap(
                check_component_status,
                [(comp, args.ignore_permissions) for comp in allcomps],
            )
        print_status(allcomps, result, max_width, args.nocolor, args.hashes)


def check_component_status(comp, ignore_permissions):
    """Check the status of a single component"""
    git = GitRepository(comp.remote, comp.local)

    # Older mepo clones will not have ignore_submodules in comp, so
    # we need to handle this gracefully
    try:
        _ignore_submodules = comp.ignore_submodules
    except AttributeError:
        _ignore_submodules = None

    # version_to_string can strip off 'origin/' for display purposes
    # so we save the "internal" name for comparison
    internal_state_branch_name = git.get_version()[0]

    # This can return non "origin/" names for detached head branches
    curr_ver = version_to_string(git.get_version(), git)
    orig_ver = version_to_string(comp.version, git)

    # This command is to try and work with git tag oddities
    curr_ver = sanitize_version_string(orig_ver, curr_ver, git)

    # We also want to see if there are any stashes in the component
    num_stashes = len(git.list_stash().splitlines())

    return (
        curr_ver,
        internal_state_branch_name,
        num_stashes,
        git.check_status(ignore_permissions, _ignore_submodules),
    )


def print_status(allcomps, result, max_width, nocolor=False, hashes=False):
    """Print the status of all components"""
    for index, comp in enumerate(allcomps):
        time.sleep(0.025)
        print_component_status(comp, result[index], max_width, nocolor, hashes)


def print_component_status(comp, result, width, nocolor=False, hashes=False):
    """Print the status of a single component"""
    current_version, internal_state_branch_name, num_stashes, output = result
    if hashes:
        comp_path = _get_relative_path(comp.local)
        comp_hash = shellcmd.run(
            cmd=shlex.split(f"git -C {comp_path} rev-parse HEAD"), output=True
        ).replace("\n", "")
        current_version = f"{current_version} ({comp_hash})"
    # This should handle tag weirdness...
    if current_version.split()[1] == comp.version.name:
        component_name = comp.name
    # Check to see if the current tag/branch is the same as the
    # original... if the above check didn't succeed, we are
    # different and we colorize if asked for
    elif (internal_state_branch_name not in comp.version.name) and not nocolor:
        component_name = colors.RED + comp.name + colors.RESET
        width += len(colors.RED) + len(colors.RESET)
    else:
        component_name = comp.name

    # If there are stashes, we print the number of stashes in yellow
    if num_stashes:
        stash_str = colors.YELLOW + f"[stashes: {num_stashes}]" + colors.RESET
        print(f"{component_name:<{width}} | {current_version} {stash_str}")
    else:
        print(f"{component_name:<{width}} | {current_version}")
    if output:
        for line in output.split("\n"):
            print("   |", line.rstrip())



import os

from shutil import get_terminal_size

from ..state import MepoState
from ..git import GitRepository
from ..utilities import verify


def run(args):
    foundDiff = False

    allcomps = MepoState.read_state()
    comps2diff = _get_comps_to_diff(args.comp_name, allcomps)

    for comp in comps2diff:
        result = check_component_diff(comp, args)
        if result:
            if not foundDiff:
                print("Diffing...", flush=True)
                foundDiff = True
            print_diff(comp, args, result)

    if not foundDiff:
        print("No diffs found")


def _get_comps_to_diff(specified_comps, allcomps):
    comps_to_diff = allcomps
    if specified_comps:
        verify.valid_components(specified_comps, allcomps)
        comps_to_diff = [x for x in allcomps if x.name in specified_comps]
    return comps_to_diff


def check_component_diff(comp, args):
    git = GitRepository(comp.remote, comp.local)

    # Older mepo clones will not have ignore_submodules in comp, so
    # we need to handle this gracefully
    try:
        _ignore_submodules = comp.ignore_submodules
    except AttributeError:
        _ignore_submodules = None
    return git.run_diff(args, _ignore_submodules)


def print_diff(comp, args, output):
    columns, lines = get_terminal_size(fallback=(80, 20))
    horiz_line = "\u2500" * columns

    print("{} (location: {}):".format(comp.name, _get_relative_path(comp.local)))
    print()
    for line in output.split("\n"):
        # print('   |', line.rstrip())
        print(line.rstrip())
    print(horiz_line)


def _get_relative_path(local_path):
    """
    Get the relative path when given a local path.

    local_path: The path to a subrepo as known by mepo (relative to the .mepo directory)
    """

    # This creates a full path on the disk from the root of mepo and the local_path
    full_local_path = os.path.join(MepoState.get_root_dir(), local_path)

    # We return the path relative to where we currently are
    return os.path.relpath(full_local_path, os.getcwd())



import argparse
import warnings

from .branch_parser import MepoBranchArgParser
from .stash_parser import MepoStashArgParser
from .tag_parser import MepoTagArgParser
from .config_parser import MepoConfigArgParser

from ..utilities import mepoconfig


def get_version():
    from importlib import metadata

    return metadata.version("mepo")


class LocationAction(argparse._StoreTrueAction):

    def __init__(self, option_strings, dest, const=True, help=None):
        super().__init__(option_strings, dest, const, help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        import os, sys
        import mepo

        print(os.path.dirname(mepo.__file__))
        sys.exit(0)


class MepoArgParser:

    __slots__ = ["parser", "subparsers"]

    def __init__(self):
        self.parser = argparse.ArgumentParser(
            description="Tool to manage (m)ultiple r(epo)s"
        )
        self.parser.add_argument("--version", action="version", version=get_version())
        self.parser.add_argument(
            "--location", action=LocationAction, help=argparse.SUPPRESS
        )
        self.subparsers = self.parser.add_subparsers()
        self.subparsers.title = "mepo commands"
        self.subparsers.required = True
        self.subparsers.dest = "mepo_cmd"

    def parse(self):
        self.__init()
        self.__clone()
        self.__list()
        self.__status()
        self.__restore_state()
        self.__diff()
        self.__fetch()
        self.__checkout()
        self.__checkout_if_exists()
        self.__changed_files()
        self.__branch()
        self.__tag()
        self.__stash()
        self.__develop()
        self.__pull()
        self.__pull_all()
        self.__compare()
        self.__reset()
        self.__whereis()
        self.__stage()
        self.__unstage()
        self.__commit()
        self.__push()
        self.__save()
        self.__config()
        self.__update_state()
        return self.parser.parse_args()

    def __init(self):
        warnings.warn(
            "init will be removed in version 3, use clone instead", DeprecationWarning
        )
        init = self.subparsers.add_parser(
            "init",
            description="Initialize mepo based on `config-file`",
            aliases=mepoconfig.get_command_alias("init"),
        )
        init.add_argument(
            "--registry",
            nargs="?",
            default="components.yaml",
            help="default: %(default)s",
        )
        init.add_argument(
            "--style",
            metavar="style-type",
            nargs="?",
            default=None,
            choices=["naked", "prefix", "postfix"],
            help="Style of directory file, default: prefix, allowed options: %(choices)s",
        )

    def __clone(self):
        clone = self.subparsers.add_parser(
            "clone",
            description="Clone repositories.",
            aliases=mepoconfig.get_command_alias("clone"),
        )
        clone.add_argument(
            "url", metavar="URL", nargs="?", default=None, help="URL to clone"
        )
        clone.add_argument(
            "directory",
            nargs="?",
            default=None,
            help="Directory to clone into (Only allowed with URL!)",
        )
        clone.add_argument(
            "--branch",
            "-b",
            metavar="name",
            nargs="?",
            default=None,
            help="Branch/tag of URL to initially clone (Only allowed with URL!)",
        )
        clone.add_argument(
            "--registry",
            metavar="registry",
            nargs="?",
            default=None,
            help="Registry (default: %(default)s)",
        )
        clone.add_argument(
            "--style",
            metavar="style-type",
            nargs="?",
            default=None,
            choices=["naked", "prefix", "postfix"],
            help="Style of directory file, default: prefix, allowed options: %(choices)s (ignored if init already called)",
        )
        clone.add_argument(
            "--allrepos",
            action="store_true",
            help="Must be passed with -b/--branch. When set, it not only checkouts out the branch/tag for the fixture, but for all the subrepositories as well.",
        )
        clone.add_argument(
            "--partial",
            metavar="partial-type",
            nargs="?",
            default=None,
            choices=[None, "blobless", "treeless"],
            help=(
                """
                Style of partial clone, default: %(default)s.
                Allowed options: %(choices)s.
                None: normal full git clone,
                blobless: cloning with "--filter=blob:none",
                treeless: cloning with "--filter=tree:0".
                NOTE: We do *not* recommend using "treeless" as it is very
                aggressive and will cause problems with many git commands.
                """
            ),
        )

    def __list(self):
        listcomps = self.subparsers.add_parser(
            "list",
            description="List all components that are being tracked",
            aliases=mepoconfig.get_command_alias("list"),
        )
        listcomps.add_argument(
            "-1", "--one-per-line", action="store_true", help="one component per line"
        )

    def __status(self):
        status = self.subparsers.add_parser(
            "status",
            description="Check current status of all components",
            aliases=mepoconfig.get_command_alias("status"),
        )
        status.add_argument(
            "--ignore-permissions",
            action="store_true",
            help="Tells command to ignore changes in file permissions.",
        )
        status.add_argument(
            "--nocolor", action="store_true", help="Tells status to not display colors."
        )
        status.add_argument(
            "--hashes", action="store_true", help="Print the exact hash of the HEAD."
        )
        status.add_argument(
            "--serial", action="store_true", help="Run the serial version."
        )

    def __restore_state(self):
        restore_state = self.subparsers.add_parser(
            "restore-state",
            description="Restores all components to the last saved state.",
            aliases=mepoconfig.get_command_alias("restore-state"),
        )
        restore_state.add_argument(
            "--serial", action="store_true", help="Run the serial version."
        )

    def __diff(self):
        diff = self.subparsers.add_parser(
            "diff",
            description="Diff all components",
            aliases=mepoconfig.get_command_alias("diff"),
        )
        diff.add_argument(
            "--name-only", action="store_true", help="Show only names of changed files"
        )
        diff.add_argument(
            "--name-status",
            action="store_true",
            help="Show name-status of changed files",
        )
        diff.add_argument(
            "--ignore-permissions",
            action="store_true",
            help="Tells command to ignore changes in file permissions.",
        )
        diff.add_argument(
            "--staged", action="store_true", help="Show diff of staged changes"
        )
        diff.add_argument(
            "-b",
            "--ignore-space-change",
            action="store_true",
            help="Ignore changes in amount of whitespace",
        )
        diff.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="Component to list branches in",
        )

    def __checkout(self):
        checkout = self.subparsers.add_parser(
            "checkout",
            description="Switch to branch/tag `branch-name` in component `comp-name`. "
            "If no components listed, checkout from all. "
            "Specifying `-b` causes the branch `branch-name` to be created and checked out.",
            aliases=mepoconfig.get_command_alias("checkout"),
        )
        checkout.add_argument(
            "branch_name", metavar="branch-name", help="Name of branch"
        )
        checkout.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="Components to checkout branch in",
        )
        checkout.add_argument("-b", action="store_true", help="create the branch")
        checkout.add_argument(
            "-q", "--quiet", action="store_true", help="Suppress prints"
        )
        checkout.add_argument(
            "--detach", action="store_true", help="Detach upon checkout"
        )

    def __checkout_if_exists(self):
        checkout_if_exists = self.subparsers.add_parser(
            "checkout-if-exists",
            description="Switch to branch or tag `ref-name` in any component where it is present. ",
            aliases=mepoconfig.get_command_alias("checkout-if-exists"),
        )
        checkout_if_exists.add_argument(
            "ref_name", metavar="ref-name", help="Name of branch or tag"
        )
        checkout_if_exists.add_argument(
            "-q", "--quiet", action="store_true", help="Suppress prints"
        )
        checkout_if_exists.add_argument(
            "--detach", action="store_true", help="Detach on checkout"
        )
        checkout_if_exists.add_argument(
            "-n",
            "--dry-run",
            action="store_true",
            help="Dry-run only (lists repos where branch exists)",
        )

    def __changed_files(self):
        changed_files = self.subparsers.add_parser(
            "changed-files",
            description="List files that have changes versus the state. By default runs against all components.",
            aliases=mepoconfig.get_command_alias("changed-files"),
        )
        changed_files.add_argument(
            "--full-path", action="store_true", help="Print with full path"
        )
        changed_files.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="Component to list branches in",
        )

    def __fetch(self):
        fetch = self.subparsers.add_parser(
            "fetch",
            description="Download objects and refs from in component `comp-name`. "
            "If no components listed, fetches from all",
            aliases=mepoconfig.get_command_alias("fetch"),
        )
        fetch.add_argument(
            "comp_name", metavar="comp-name", nargs="*", help="Components to fetch in"
        )
        fetch.add_argument("--all", action="store_true", help="Fetch all remotes.")
        fetch.add_argument(
            "-p", "--prune", action="store_true", help="Prune remote branches."
        )
        fetch.add_argument("-t", "--tags", action="store_true", help="Fetch tags.")
        fetch.add_argument("-f", "--force", action="store_true", help="Force action.")

    def __branch(self):
        branch = self.subparsers.add_parser(
            "branch",
            description="Runs branch commands.",
            aliases=mepoconfig.get_command_alias("branch"),
        )
        MepoBranchArgParser(branch)

    def __stash(self):
        stash = self.subparsers.add_parser(
            "stash",
            description="Runs stash commands.",
            aliases=mepoconfig.get_command_alias("stash"),
        )
        MepoStashArgParser(stash)

    def __tag(self):
        tag = self.subparsers.add_parser(
            "tag",
            description="Runs tag commands.",
            aliases=mepoconfig.get_command_alias("tag"),
        )
        MepoTagArgParser(tag)

    def __develop(self):
        develop = self.subparsers.add_parser(
            "develop",
            description="Checkout current version of 'develop' branches of specified components",
            aliases=mepoconfig.get_command_alias("develop"),
        )
        develop.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            default=None,
            help="Component(s) to checkout development branches",
        )
        develop.add_argument(
            "-q", "--quiet", action="store_true", help="Suppress prints"
        )

    def __pull(self):
        pull = self.subparsers.add_parser(
            "pull",
            description="Pull branches of specified components",
            aliases=mepoconfig.get_command_alias("pull"),
        )
        pull.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            default=None,
            help="Components to pull in",
        )
        pull.add_argument("-q", "--quiet", action="store_true", help="Suppress prints")

    def __pull_all(self):
        pull_all = self.subparsers.add_parser(
            "pull-all",
            description="Pull branches of all components (only those in non-detached HEAD state)",
            aliases=mepoconfig.get_command_alias("pull-all"),
        )
        pull_all.add_argument(
            "-q", "--quiet", action="store_true", help="Suppress prints"
        )

    def __compare(self):
        compare = self.subparsers.add_parser(
            "compare",
            description="Compare current and original states of all components. "
            "Will only show differing repos unless --all is passed in",
            aliases=mepoconfig.get_command_alias("compare"),
        )
        compare.add_argument(
            "--all",
            action="store_true",
            help="Show all repos, not only differing repos",
        )
        compare.add_argument(
            "--nocolor",
            action="store_true",
            help="Tells command to not display colors.",
        )
        compare.add_argument(
            "--wrap",
            action="store_true",
            help="Tells command to ignore terminal size and wrap",
        )

    def __reset(self):
        reset = self.subparsers.add_parser(
            "reset",
            description="Reset the current mepo clone to the original state. "
            "This will delete all subrepos and does not check for uncommitted changes! "
            "Must be run in the root of the mepo clone.",
            aliases=mepoconfig.get_command_alias("reset"),
        )
        reset.add_argument("-f", "--force", action="store_true", help="Force action.")
        reset.add_argument(
            "--reclone", action="store_true", help="Reclone repos after reset."
        )
        reset.add_argument("-n", "--dry-run", action="store_true", help="Dry-run only")

    def __whereis(self):
        whereis = self.subparsers.add_parser(
            "whereis",
            description="Get the location of component `comp-name` "
            "relative to my current location. If `comp-name` is not present, "
            "get the relative locations of ALL components.",
            aliases=mepoconfig.get_command_alias("whereis"),
        )
        whereis.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="?",
            default=None,
            help="Component to get location of",
        )
        whereis.add_argument(
            "-i", "--ignore-case", action="store_true", help="Ignore case for whereis"
        )

    def __stage(self):
        stage = self.subparsers.add_parser(
            "stage",
            description="Stage modified & untracked files in the specified component(s)",
            aliases=mepoconfig.get_command_alias("stage"),
        )
        stage.add_argument(
            "--untracked", action="store_true", help="Stage untracked files as well"
        )
        stage.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="Component to stage file in",
        )

    def __unstage(self):
        unstage = self.subparsers.add_parser(
            "unstage",
            description="Un-stage staged files. "
            "If a component is specified, files are un-staged only for that component.",
            aliases=mepoconfig.get_command_alias("unstage"),
        )
        unstage.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="*",
            help="Component to unstage in",
            default=None,
        )

    def __commit(self):
        commit = self.subparsers.add_parser(
            "commit",
            description="Commit staged files in the specified components",
            aliases=mepoconfig.get_command_alias("commit"),
        )
        commit.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="Stage all tracked files and then commit",
        )
        commit.add_argument(
            "-m",
            "--message",
            type=str,
            metavar="message",
            default=None,
            help="Message to commit with",
        )
        commit.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="Component to commit file in",
        )

    def __push(self):
        push = self.subparsers.add_parser(
            "push",
            description="Push local commits to remote for specified component. "
            "Use mepo tag push to push tags",
            aliases=mepoconfig.get_command_alias("push"),
        )
        push.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="Component to push to remote",
        )

    def __save(self):
        save = self.subparsers.add_parser(
            "save",
            description="Save current state in a yaml registry",
            aliases=mepoconfig.get_command_alias("save"),
        )
        save.add_argument(
            "registry",
            metavar="registry",
            nargs="?",
            default="components-new.yaml",
            help="default: %(default)s",
        )

    def __config(self):
        config = self.subparsers.add_parser(
            "config",
            description="Runs config commands.",
            aliases=mepoconfig.get_command_alias("config"),
        )
        MepoConfigArgParser(config)

    def __update_state(self):
        _ = self.subparsers.add_parser(
            "update-state",
            description="Permanently update mepo1 state to current",
            aliases=mepoconfig.get_command_alias("update-state"),
        )



import os
import tempfile
import subprocess

from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository
from ..git import get_editor as get_git_editor

from .stage import stage_files


def run(args):
    allcomps = MepoState.read_state()
    verify.valid_components(args.comp_name, allcomps)
    comps2commit = [x for x in allcomps if x.name in args.comp_name]

    tf_file = None

    # Pop up an editor if a message is not provided
    # Popping up an EDITOR is based on https://stackoverflow.com/a/39989442
    if not args.message:
        EDITOR = get_git_editor()
        initial_message = b""  # set up the file

        # Use delete=False to keep the file around as we send the file name to git commit -F
        tf = tempfile.NamedTemporaryFile(delete=False)
        tf_file = tf.name
        tf.write(initial_message)
        tf.flush()
        subprocess.call([EDITOR, tf.name])

    for comp in comps2commit:
        git = GitRepository(comp.remote, comp.local)
        if args.all:
            stage_files(git, comp, commit=True)

        staged_files = git.get_staged_files()
        if staged_files:
            git.commit_files(args.message, tf_file)

        for myfile in staged_files:
            print("+ {}: {}".format(comp.name, myfile))

    # Now close and by-hand delete the temp file
    if not args.message:
        tf.close()
        os.unlink(tf.name)



class MepoConfigArgParser:

    def __init__(self, config):
        self.config = config.add_subparsers()
        self.config.title = "mepo config sub-commands"
        self.config.dest = "mepo_config_cmd"
        self.config.required = True
        self.__get()
        self.__set()
        self.__delete()
        self.__print()

    def __get(self):
        get = self.config.add_parser(
            "get",
            description=(
                "Get config `entry` in `.mepoconfig`. "
                "Note this uses gitconfig style where `entry` is of the form `section.option`. "
                "So to get an `alias` `st` You would run `mepo config get alias.st`"
            ),
        )
        get.add_argument("entry", metavar="entry", help="Entry to display.")

    def __set(self):
        set_ = self.config.add_parser(
            "set",
            description=(
                "Set config `entry` to `value` in `.mepoconfig`. "
                "Note this uses gitconfig style where `entry` is of the form `section.option`. "
                "So to set an `alias` for `status` of `st` You would run `mepo config set alias.st status`"
            ),
        )
        set_.add_argument("entry", metavar="entry", help="Entry to set.")
        set_.add_argument("value", metavar="value", help="Value to set entry to.")

    def __delete(self):
        delete = self.config.add_parser(
            "delete",
            description=(
                "Delete config `entry` in `.mepoconfig`. "
                "Note this uses gitconfig style where `entry` is of the form `section.option`. "
                "So to delete an `alias` `st` You would run `mepo config delete alias.st`"
            ),
        )
        delete.add_argument("entry", metavar="entry", help="Entry to delete.")

    def __print(self):
        _ = self.config.add_parser(
            "print", description="Print contents of `.mepoconfig`"
        )



from ..state import MepoState
from ..utilities import mepoconfig


def run(args):
    if args.style:
        style = args.style
    elif mepoconfig.has_option("init", "style"):
        allowed_styles = ["naked", "prefix", "postfix"]
        style = mepoconfig.get("init", "style")
        if style not in allowed_styles:
            raise Exception(
                f"Detected style [{style}] from .mepoconfig is not an allowed style: {allowed_styles}"
            )
        else:
            print(f"Found style [{style}] in .mepoconfig")
    else:
        style = None

    _ = MepoState.initialize(args.registry, style)

    if not style:
        print(f"Initializing mepo using {args.registry}")
    else:
        print(f"Initializing mepo using {args.registry} with {style} style")



from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository
from ..component import MepoVersion


def run(args):
    allcomps = MepoState.read_state()
    verify.valid_components(args.comp_name, allcomps)
    comps2stg = [x for x in allcomps if x.name in args.comp_name]
    for comp in comps2stg:
        git = GitRepository(comp.remote, comp.local)
        stage_files(git, comp, args.untracked)


def stage_files(git, comp, untracked=False, commit=False):
    curr_ver = MepoVersion(*git.get_version())
    if curr_ver.detached:  # detached head
        raise Exception(f"{comp.name} has detached head! Cannot stage.")
    for myfile in git.get_changed_files(untracked=untracked):
        git.stage_file(myfile)
        print_output = f"{comp.name}: {myfile}"
        if commit:
            print(f"Staged: {print_output}")
        else:
            print(f"+ {print_output}")



class MepoStashArgParser:

    def __init__(self, stash):
        self.stash = stash.add_subparsers()
        self.stash.title = "mepo stash sub-commands"
        self.stash.dest = "mepo_stash_cmd"
        self.stash.required = True
        self.__push()
        self.__list()
        self.__pop()
        self.__apply()
        self.__show()

    def __push(self):
        stpush = self.stash.add_parser(
            "push", description="Push (create) stash in component <comp-name>"
        )
        stpush.add_argument(
            "-m",
            "--message",
            type=str,
            metavar="message",
            default=None,
            help="Message for the stash",
        )
        stpush.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="Component to push stash in",
        )

    def __show(self):
        stshow = self.stash.add_parser(
            "show", description="show stash in component <comp-name>"
        )
        stshow.add_argument(
            "-p", "--patch", action="store_true", help="Message for the stash"
        )
        stshow.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="Component to show stash in",
        )

    def __list(self):
        _ = self.stash.add_parser(
            "list", description="List local stashes of all components"
        )

    def __pop(self):
        stpop = self.stash.add_parser(
            "pop", description="Pop stash in component <comp-name>"
        )
        stpop.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="Component to pop stash in",
        )

    def __apply(self):
        stapply = self.stash.add_parser(
            "apply", description="apply stash in component <comp-name>"
        )
        stapply.add_argument(
            "comp_name",
            metavar="comp-name",
            nargs="+",
            help="Component to apply stash in",
        )



from ..state import MepoState
from ..utilities import verify
from ..utilities import colors
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    comps2checkout = _get_comps_to_checkout(args.comp_name, allcomps)
    for comp in comps2checkout:
        git = GitRepository(comp.remote, comp.local)
        branch = args.branch_name
        if args.b:
            git.create_branch(branch)
            if not args.quiet:
                # print('+ {}: {}'.format(comp.name, branch))
                print(
                    "Creating and checking out branch %s in %s"
                    % (
                        colors.YELLOW + branch + colors.RESET,
                        colors.RESET + comp.name + colors.RESET,
                    )
                )
        else:
            if not args.quiet:
                print(
                    "Checking out %s in %s"
                    % (
                        colors.YELLOW + branch + colors.RESET,
                        colors.RESET + comp.name + colors.RESET,
                    )
                )
        git.checkout(branch, args.detach)


def _get_comps_to_checkout(specified_comps, allcomps):
    comps_to_list = allcomps
    if specified_comps:
        verify.valid_components(specified_comps, allcomps)
        comps_to_list = [x for x in allcomps if x.name in specified_comps]
    return comps_to_list



from ..utilities import verify
from ..state import MepoState
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    verify.valid_components(args.comp_name, allcomps)
    comps2push = [x for x in allcomps if x.name in args.comp_name]
    for comp in comps2push:
        git = GitRepository(comp.remote, comp.local)
        output = git.push()
        print("----------\nPushed: {}\n----------".format(comp.name))
        print(output)



import subprocess as sp


def run(cmd, output=None, stdout=None, status=None):
    result = sp.run(
        cmd,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True,  # result byte sequence -> string
    )

    if status:
        return result.returncode
    elif result.returncode != 0:
        print(result.stderr)
        result.check_returncode()

    if stdout:
        return result.stdout
    if output:
        return result.stdout + result.stderr



from shutil import get_terminal_size

from ..state import MepoState

from ..utilities import colors
from ..utilities.version import version_to_string
from ..utilities.version import sanitize_version_string

from ..git import GitRepository

VER_LEN = 30


def run(args):
    allcomps = MepoState.read_state()

    if not any_differing_repos(allcomps):
        print("No repositories have changed")
    else:
        max_namelen, max_origlen = calculate_header_lengths(allcomps, args.all)
        print_header(max_namelen, max_origlen)
        for comp in allcomps:
            git = GitRepository(comp.remote, comp.local)
            curr_ver = version_to_string(git.get_version(), git)
            orig_ver = version_to_string(comp.version, git)

            # This command is to try and work with git tag oddities
            curr_ver = sanitize_version_string(orig_ver, curr_ver, git)

            print_cmp(
                comp.name,
                orig_ver,
                curr_ver,
                max_namelen,
                max_origlen,
                args.all,
                args.nocolor,
                args.wrap,
            )


def any_differing_repos(allcomps):

    for comp in allcomps:
        git = GitRepository(comp.remote, comp.local)
        curr_ver = version_to_string(git.get_version(), git)
        orig_ver = version_to_string(comp.version, git)

        # This command is to try and work with git tag oddities
        curr_ver = sanitize_version_string(orig_ver, curr_ver, git)

        if curr_ver not in orig_ver:
            return True

    return False


def calculate_header_lengths(allcomps, all_repos=False):
    names = []
    versions = []
    for comp in allcomps:
        git = GitRepository(comp.remote, comp.local)

        # We want to base the display on changed repos
        # if we don't ask for all repos
        if not all_repos:
            curr_ver = version_to_string(git.get_version(), git)
            orig_ver = version_to_string(comp.version, git)

            # This command is to try and work with git tag oddities
            curr_ver = sanitize_version_string(orig_ver, curr_ver, git)

            if curr_ver not in orig_ver:
                names.append(comp.name)
                versions.append(version_to_string(comp.version, git))
        else:
            names.append(comp.name)
            versions.append(version_to_string(comp.version, git))
    max_namelen = len(max(names, key=len))
    # Note: max_namelen could be 3 characters but we want at least 4 for "Repo"
    max_namelen = max(max_namelen, 4)
    max_origlen = len(max(versions, key=len))
    return max_namelen, max_origlen


def print_header(max_namelen, max_origlen):
    FMT_VAL = (max_namelen, max_namelen, max_origlen)
    FMTHEAD = "{:<%s.%ss} | {:<%ss} | {:<s}" % FMT_VAL
    print(FMTHEAD.format("Repo", "Original", "Current"))
    print(FMTHEAD.format("-" * 80, "-" * max_origlen, "-" * 7))


def print_cmp(
    name, orig, curr, name_width, orig_width, all_repos=False, nocolor=False, wrap=False
):
    name_blank = ""
    # if orig not in curr:
    if curr not in orig:
        if not nocolor:
            name = colors.RED + name + colors.RESET
            name_blank = colors.RED + name_blank + colors.RESET
            name_width += len(colors.RED) + len(colors.RESET)
    else:
        # This only prints differing repos unless --all is passed in
        if not all_repos:
            return
    FMT_VAL = (name_width, name_width, orig_width)

    FMT0 = "{:<%s.%ss} | {:<%ss} | {:<s}" % FMT_VAL
    FMT1 = "{:<%s.%ss} | {:<%ss}" % FMT_VAL
    FMT2 = "{:<%s.%ss} | {:>%ss} | {:<s}" % FMT_VAL

    columns, lines = get_terminal_size(fallback=(80, 20))

    if (not wrap) and (len(FMT0.format(name, orig, curr)) > columns):
        print(FMT1.format(name, orig + " ..."))
        print(FMT2.format(name_blank, "...", curr))
    else:
        print(FMT0.format(name, orig, curr))



from ..state import MepoState
from ..component import MepoVersion
from ..utilities import verify
from ..utilities import colors
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    verify.valid_components(args.comp_name, allcomps)
    comps2pull = [x for x in allcomps if x.name in args.comp_name]
    for comp in comps2pull:
        git = GitRepository(comp.remote, comp.local)
        name, tYpe, is_detached = MepoVersion(*git.get_version())
        if is_detached:
            raise Exception("{} has detached head! Cannot pull.".format(comp.name))
        else:
            print(
                "Pulling branch %s in %s "
                % (
                    colors.YELLOW + name + colors.RESET,
                    colors.RESET + comp.name + colors.RESET,
                )
            )
            output = git.pull()
            if not args.quiet:
                print(output)



"""Permanently convert mepo1 state to mepo2 state"""

from urllib.parse import urljoin

from ..state import MepoState
from ..git import get_current_remote_url


def run(_):
    """Entry point"""
    if __old_style_exists_but_new_style_does_not():
        # mepo2 style does not exist
        allcomps = MepoState.read_state()
        MepoState.mepo1_patch_undo()
        for comp in allcomps:
            if comp.remote.startswith("../"):
                comp.remote = urljoin(get_current_remote_url() + "/", comp.remote)
        # Write new state
        MepoState.write_state(allcomps)
        print("\nConverted mepo1 state to mepo2\n")
    else:
        print("Detected mepo2 style state - nothing to do")


def __old_style_exists_but_new_style_does_not():
    if MepoState.state_exists(old_style=True):
        if not MepoState.state_exists(old_style=False):
            return True
    return False



import os
import shutil
import pathlib
from urllib.parse import urlparse
from types import SimpleNamespace

from .init import run as mepo_init
from ..state import MepoState
from ..state import StateDoesNotExistError
from ..git import GitRepository
from ..utilities import colors
from ..utilities import mepoconfig


def run(args):
    """
    Entry point of clone.

    Multiple ways to run clone
    1. After fixture has been cloned (via git clone)
       a. mepo init
          mepo clone
       b. mepo clone (initializes mepo)
    2. Clone fixture as well
       a. mepo clone <url> [<directory>]
       b. mepo clone -b <branch> <url> [<directory>]

    Steps -
    1. Clone fixture - if url is provided
    2. Read state - initialize mepo (write state) first, if needed
    3. Clone components
    4. Checkout all repos to the specified branch
    """
    CWD = os.getcwd()

    arg_partial = handle_partial(args.partial)

    if args.url is not None:
        fixture_dir = clone_fixture(args.url, args.branch, args.directory, arg_partial)
        os.chdir(fixture_dir)
    allcomps = read_state(args.style, args.registry)
    clone_components(allcomps, arg_partial)
    if args.allrepos:
        checkout_all_repos(allcomps, args.branch)

    os.chdir(CWD)


def handle_partial(partial):
    """
    The `partial` argument to clone can be set either via command line or
    via .mepoconfig. Non-default value set via command line takes precedence.
    The default value of `partial` is None, and possible choices are None/blobless/treeless
    """
    ALLOWED_NON_DEFAULT = ["blobless", "treeless"]
    if partial is None:  # default value from command line
        if mepoconfig.has_option("clone", "partial"):
            partial = mepoconfig.get("clone", "partial")
            if partial not in ALLOWED_NON_DEFAULT:
                raise ValueError(f"Invalid partial type [{partial}] in .mepoconfig")
            print(f"Found partial clone type [{partial}] in .mepoconfig")
    return partial


def clone_fixture(url, branch=None, directory=None, partial=None):
    if directory is None:
        p = urlparse(url)
        last_url_node = p.path.rsplit("/")[-1]
        directory = pathlib.Path(last_url_node).stem
    git = GitRepository(url, directory)
    git.clone(branch, partial)
    return directory


def read_state(arg_style, arg_registry):
    while True:
        try:
            allcomps = MepoState.read_state()
        except StateDoesNotExistError:
            registry = get_registry(arg_registry)
            mepo_init(SimpleNamespace(style=arg_style, registry=registry))
            continue
        break
    return allcomps


def get_registry(arg_registry):
    registry = "components.yaml"
    if arg_registry is not None:
        shutil.copy(arg_registry, os.getcwd())
        registry = os.path.basename(arg_registry)
    return registry


def clone_components(allcomps, partial):
    max_namelen = max([len(comp.name) for comp in allcomps])
    for comp in allcomps:
        if comp.fixture:
            continue  # not cloning fixture
        recurse_submodules = comp.recurse_submodules
        # According to Git, treeless clones do not interact well with
        # submodules. So if any comp has the recurse option set to True,
        # we do a non-partial clone
        partial = None if partial == "treeless" and recurse_submodules else partial
        version = comp.version.name
        version = version.replace("origin/", "")
        git = GitRepository(comp.remote, comp.local)
        git.clone(version, recurse_submodules, partial)
        if comp.sparse:
            git.sparsify(comp.sparse)
        print_clone_info(comp.name, comp.version, max_namelen)


def print_clone_info(comp_name, comp_version, name_width):
    ver_name_type = f"({comp_version.type}) {comp_version.name}"
    print(f"{comp_name:<{name_width}} | {ver_name_type:<s}")


def checkout_all_repos(allcomps, branch):
    if branch is None:
        raise RuntimeError("`allrepos` option must be used with a branch/tag.")
    for comp in allcomps:
        branch_y = colors.YELLOW + branch + colors.RESET
        print(f"Checking out {branch_y} in {comp.name}")
        git = GitRepository(comp.remote, comp.local)
        git.checkout(branch)



import os
import shutil
import shlex
import subprocess as sp

from urllib.parse import urljoin

from .utilities import shellcmd
from .utilities import colors
from .utilities.exceptions import RepoAlreadyClonedError


def get_editor():
    """
    Return GIT_EDITOR
    """
    result = sp.run(
        "git var GIT_EDITOR".split(), stdout=sp.PIPE, stderr=sp.PIPE, check=True
    )
    return result.stdout.rstrip().decode("utf-8")  # byte to utf-8


def get_current_remote_url():
    cmd = "git remote get-url origin"
    output = shellcmd.run(shlex.split(cmd), output=True).strip()
    return output


class GitRepository:
    """
    Class to consolidate git commands
    """

    __slots__ = ["__local_path_abs", "__remote", "__git"]

    def __init__(self, remote_url, local_path_abs):
        self.__local_path_abs = local_path_abs
        self.__remote = remote_url
        self.__git = 'git -C "{}"'.format(self.__local_path_abs)

    def get_local_path(self):
        return self.__local_path_abs

    def get_remote_url(self):
        return self.__remote

    def clone(self, version=None, recurse=None, partial=None):
        """
        Execute `git clone` command
        version is tag or branch
        """
        PARTIAL = {"blobless": " --filter=blob:none", "treeless": " --filter=tree:0"}

        cmd = "git clone "
        if partial is not None:
            cmd += PARTIAL[partial]
        if recurse is not None:
            cmd += " --recurse-submodules "
        cmd += " --quiet {} {}".format(self.__remote, self.__local_path_abs)
        shellcmd.run(shlex.split(cmd))

        if version is not None:
            self.checkout(version)

    def checkout(self, version, detach=False):
        cmd = self.__git + " checkout "
        cmd += "--quiet {}".format(version)
        shellcmd.run(shlex.split(cmd))
        if detach:
            cmd2 = self.__git + " checkout --detach"
            shellcmd.run(shlex.split(cmd2))

    def sparsify(self, sparse_config):
        dst = os.path.join(self.__local_path_abs, ".git", "info", "sparse-checkout")
        os.makedirs(os.path.dirname(dst), exist_ok=True)
        shutil.copy(sparse_config, dst)
        cmd1 = self.__git + " config core.sparseCheckout true"
        shellcmd.run(shlex.split(cmd1))
        cmd2 = self.__git + " read-tree -mu HEAD"
        shellcmd.run(shlex.split(cmd2))

    def list_branch(self, all=False, nocolor=False):
        cmd = self.__git + " branch"
        if all:
            cmd += " -a"
        if nocolor:
            cmd += " --color=never"
        return shellcmd.run(shlex.split(cmd), output=True)

    def list_tags(self):
        cmd = self.__git + " tag"
        return shellcmd.run(shlex.split(cmd), output=True)

    def rev_list(self, tag):
        cmd = self.__git + " rev-list -n 1 {}".format(tag)
        return shellcmd.run(shlex.split(cmd), output=True)

    def rev_parse(self, short=False):
        cmd = self.__git + " rev-parse --verify HEAD"
        if short:
            cmd += " --short"
        return shellcmd.run(shlex.split(cmd), output=True)

    def list_stash(self):
        cmd = self.__git + " stash list"
        return shellcmd.run(shlex.split(cmd), output=True)

    def pop_stash(self):
        cmd = self.__git + " stash pop"
        return shellcmd.run(shlex.split(cmd), output=True)

    def apply_stash(self):
        cmd = self.__git + " stash apply"
        return shellcmd.run(shlex.split(cmd), output=True)

    def push_stash(self, message):
        cmd = self.__git + " stash push"
        if message:
            cmd += " -m {}".format(message)
        return shellcmd.run(shlex.split(cmd), output=True)

    def show_stash(self, patch):
        cmd = self.__git + " stash show"
        if patch:
            cmd += " -p --color"
        output = shellcmd.run(shlex.split(cmd), output=True)
        return output.rstrip()

    def run_diff(self, args=None, ignore_submodules=False):
        cmd = "git -C {}".format(self.__local_path_abs)
        if args.ignore_permissions:
            cmd += " -c core.fileMode=false"
        cmd += " diff --color"
        if args.name_only:
            cmd += " --name-only"
        if args.name_status:
            cmd += " --name-status"
        if args.staged:
            cmd += " --staged"
        if args.ignore_space_change:
            cmd += " --ignore-space-change"
        if ignore_submodules:
            cmd += " --ignore-submodules=all"
        output = shellcmd.run(shlex.split(cmd), output=True)
        return output.rstrip()

    def fetch(self, args=None):
        cmd = self.__git + " fetch"
        if args.all:
            cmd += " --all"
        if args.prune:
            cmd += " --prune"
        if args.tags:
            cmd += " --tags"
        if args.force:
            cmd += " --force"
        return shellcmd.run(shlex.split(cmd), output=True)

    def create_branch(self, branch_name):
        cmd = self.__git + " branch {}".format(branch_name)
        shellcmd.run(shlex.split(cmd))

    def create_tag(self, tag_name, annotate, message, tf_file=None):
        if annotate:
            if tf_file:
                cmd = [
                    "git",
                    "-C",
                    self.__local_path_abs,
                    "tag",
                    "-a",
                    "-F",
                    tf_file,
                    tag_name,
                ]
            elif message:
                cmd = [
                    "git",
                    "-C",
                    self.__local_path_abs,
                    "tag",
                    "-a",
                    "-m",
                    message,
                    tag_name,
                ]
            else:
                raise Exception("This should not happen")
        else:
            cmd = ["git", "-C", self.__local_path_abs, "tag", tag_name]
        shellcmd.run(cmd)

    def delete_branch(self, branch_name, force):
        delete = "-d"
        if force:
            delete = "-D"
        cmd = self.__git + " branch {} {}".format(delete, branch_name)
        shellcmd.run(shlex.split(cmd))

    def delete_tag(self, tag_name):
        cmd = self.__git + " tag -d {}".format(tag_name)
        shellcmd.run(shlex.split(cmd))

    def push_tag(self, tag_name, force, delete):
        cmd = self.__git + " push"
        if force:
            cmd += " --force"
        if delete:
            cmd += " --delete"
        cmd += " origin {}".format(tag_name)
        shellcmd.run(shlex.split(cmd))

    def verify_branch_or_tag(self, ref_name):
        branch_cmd = self.__git + f" show-branch remotes/origin/{ref_name}"
        tag_cmd = self.__git + f" rev-parse {ref_name}"
        branch_status = shellcmd.run(shlex.split(branch_cmd), status=True)
        ref_type = "UNKNOWN"
        if branch_status != 0:
            status = shellcmd.run(shlex.split(tag_cmd), status=True)
            ref_type = "Tag"
        else:
            status = branch_status
            ref_type = "Branch"
        return status, ref_type

    def check_status(self, ignore_permissions=False, ignore_submodules=False):
        cmd = "git -C {}".format(self.__local_path_abs)
        if ignore_permissions:
            cmd += " -c core.fileMode=false"
        cmd += " status --porcelain=v2"
        if ignore_submodules:
            cmd += " --ignore-submodules=all"
        output = shellcmd.run(shlex.split(cmd), output=True)
        if output.strip():
            output_list = output.splitlines()

            # Grab the file names first for pretty printing
            file_name_list = [item.split()[-1] for item in output_list]
            max_file_name_length = len(max(file_name_list, key=len))

            verbose_output_list = []
            for item in output_list:

                index_field = item.split()[0]
                if index_field == "2":
                    new_file_name = colors.YELLOW + item.split()[-2] + colors.RESET

                file_name = item.split()[-1]

                short_status = item.split()[1]

                if index_field == "?":
                    verbose_status = colors.RED + "untracked file" + colors.RESET

                elif short_status == ".D":
                    verbose_status = colors.RED + "deleted, not staged" + colors.RESET
                elif short_status == ".M":
                    verbose_status = colors.RED + "modified, not staged" + colors.RESET
                elif short_status == ".A":
                    verbose_status = colors.RED + "added, not staged" + colors.RESET
                elif short_status == ".T":
                    verbose_status = (
                        colors.RED + "typechange, not staged" + colors.RESET
                    )

                elif short_status == "D.":
                    verbose_status = colors.GREEN + "deleted, staged" + colors.RESET
                elif short_status == "M.":
                    verbose_status = colors.GREEN + "modified, staged" + colors.RESET
                elif short_status == "A.":
                    verbose_status = colors.GREEN + "added, staged" + colors.RESET
                elif short_status == "T.":
                    verbose_status = colors.GREEN + "typechange, staged" + colors.RESET

                elif short_status == "MM":
                    verbose_status = (
                        colors.GREEN
                        + "modified, staged"
                        + colors.RESET
                        + " with "
                        + colors.RED
                        + "unstaged changes"
                        + colors.RESET
                    )
                elif short_status == "MD":
                    verbose_status = (
                        colors.GREEN
                        + "modified, staged"
                        + colors.RESET
                        + " but "
                        + colors.RED
                        + "deleted, not staged"
                        + colors.RESET
                    )

                elif short_status == "AM":
                    verbose_status = (
                        colors.GREEN
                        + "added, staged"
                        + colors.RESET
                        + " with "
                        + colors.RED
                        + "unstaged changes"
                        + colors.RESET
                    )
                elif short_status == "AD":
                    verbose_status = (
                        colors.GREEN
                        + "added, staged"
                        + colors.RESET
                        + " but "
                        + colors.RED
                        + "deleted, not staged"
                        + colors.RESET
                    )

                elif short_status == "TM":
                    verbose_status = (
                        colors.GREEN
                        + "typechange, staged"
                        + colors.RESET
                        + " with "
                        + colors.RED
                        + "unstaged changes"
                        + colors.RESET
                    )
                elif short_status == "TD":
                    verbose_status = (
                        colors.GREEN
                        + "typechange, staged"
                        + colors.RESET
                        + " but "
                        + colors.RED
                        + "deleted, not staged"
                        + colors.RESET
                    )

                elif short_status == "R.":
                    verbose_status = (
                        colors.GREEN
                        + "renamed"
                        + colors.RESET
                        + " as "
                        + colors.YELLOW
                        + new_file_name
                        + colors.RESET
                    )
                elif short_status == "RM":
                    verbose_status = (
                        colors.GREEN
                        + "renamed, staged"
                        + colors.RESET
                        + " as "
                        + colors.YELLOW
                        + new_file_name
                        + colors.RESET
                        + " with "
                        + colors.RED
                        + "unstaged changes"
                        + colors.RESET
                    )
                elif short_status == "RD":
                    verbose_status = (
                        colors.GREEN
                        + "renamed, staged"
                        + colors.RESET
                        + " as "
                        + colors.YELLOW
                        + new_file_name
                        + colors.RESET
                        + " but "
                        + colors.RED
                        + "deleted, not staged"
                        + colors.RESET
                    )

                elif short_status == "C.":
                    verbose_status = (
                        colors.GREEN
                        + "copied"
                        + colors.RESET
                        + " as "
                        + colors.YELLOW
                        + new_file_name
                        + colors.RESET
                    )
                elif short_status == "CM":
                    verbose_status = (
                        colors.GREEN
                        + "copied, staged"
                        + colors.RESET
                        + " as "
                        + colors.YELLOW
                        + new_file_name
                        + colors.RESET
                        + " with "
                        + colors.RED
                        + "unstaged changes"
                        + colors.RESET
                    )
                elif short_status == "CD":
                    verbose_status = (
                        colors.GREEN
                        + "copied, staged"
                        + colors.RESET
                        + " as "
                        + colors.YELLOW
                        + new_file_name
                        + colors.RESET
                        + " but "
                        + colors.RED
                        + "deleted, not staged"
                        + colors.RESET
                    )

                else:
                    verbose_status = (
                        colors.CYAN
                        + "unknown"
                        + colors.RESET
                        + " (please contact mepo maintainer)"
                    )

                verbose_status_string = (
                    "{file_name:>{file_name_length}}: {verbose_status}".format(
                        file_name=file_name,
                        file_name_length=max_file_name_length,
                        verbose_status=verbose_status,
                    )
                )
                verbose_output_list.append(verbose_status_string)

            output = "\n".join(verbose_output_list)

        return output.rstrip()

    def __get_modified_files(self, orig_ver, comp_type):
        if not orig_ver:
            cmd = self.__git + " diff --name-only"
        else:
            if comp_type == "b":
                cmd = self.__git + " diff --name-only origin/{}".format(orig_ver)
            else:
                cmd = self.__git + " diff --name-only {}".format(orig_ver)
        output = shellcmd.run(shlex.split(cmd), output=True).strip()
        return output.split("\n") if output else []

    def __get_untracked_files(self):
        cmd = self.__git + " ls-files --others --exclude-standard"
        output = shellcmd.run(shlex.split(cmd), output=True).strip()
        return output.split("\n") if output else []

    def get_changed_files(self, untracked=False, orig_ver=None, comp_type=None):
        changed_files = self.__get_modified_files(orig_ver, comp_type)
        if untracked:
            changed_files += self.__get_untracked_files()
        return changed_files

    def stage_file(self, myfile):
        cmd = self.__git + " add {}".format(myfile)
        shellcmd.run(shlex.split(cmd))

    def get_staged_files(self):
        cmd = self.__git + " diff --name-only --staged"
        output = shellcmd.run(shlex.split(cmd), output=True).strip()
        return output.split("\n") if output else []

    def unstage_file(self, myfile):
        cmd = self.__git + " reset -- {}".format(myfile)
        shellcmd.run(shlex.split(cmd))

    def commit_files(self, message, tf_file=None):
        if tf_file:
            cmd = ["git", "-C", self.__local_path_abs, "commit", "-F", tf_file]
        elif message:
            cmd = ["git", "-C", self.__local_path_abs, "commit", "-m", message]
        else:
            raise Exception("This should not happen")
        shellcmd.run(cmd)

    def push(self):
        cmd = self.__git + " push -u {}".format(self.__remote)
        return shellcmd.run(shlex.split(cmd), output=True).strip()

    def get_remote_latest_commit_id(self, branch, commit_type):
        if commit_type == "h":
            cmd = self.__git + " cat-file -e {}".format(branch)
            status = shellcmd.run(shlex.split(cmd), status=True)
            if status != 0:
                msg = "Hash {} does not exist on {}".format(branch, self.__remote)
                msg += " Have you run 'mepo push'?"
                raise RuntimeError(msg)
            return branch
        else:
            # If we are a branch...
            if commit_type == "b":
                msgtype = "Branch"
                reftype = "heads"
            elif commit_type == "t":
                msgtype = "Tag"
                reftype = "tags"
            else:
                raise RuntimeError("Should not get here")
            cmd = self.__git + " ls-remote {} refs/{}/{}".format(
                self.__remote, reftype, branch
            )
            output = shellcmd.run(shlex.split(cmd), stdout=True).strip()
            if not output:
                # msg = '{} {} does not exist on {}'.format(msgtype, branch, self.__remote)
                # msg += " Have you run 'mepo push'?"
                # raise RuntimeError(msg)
                cmd = self.__git + " rev-parse HEAD"
            output = shellcmd.run(shlex.split(cmd), output=True).strip()
            return output.split()[0]

    def get_local_latest_commit_id(self):
        cmd = self.__git + " rev-parse HEAD"
        return shellcmd.run(shlex.split(cmd), output=True).strip()

    def pull(self):
        cmd = self.__git + " pull"
        return shellcmd.run(shlex.split(cmd), output=True).strip()

    def get_version(self):
        cmd = self.__git + " show -s --pretty=%D HEAD"
        output = shellcmd.run(shlex.split(cmd), output=True)
        if output.startswith("HEAD ->"):  # an actual branch
            detached = False
            name = output.split(",")[0].split("->")[1].strip()
            tYpe = "b"
        elif output.startswith("HEAD,"):  # detached head
            detached = True
            tmp = output.split(",")[1].strip()
            if tmp.startswith("tag:"):  # tag
                name = tmp[5:]
                tYpe = "t"
            else:
                # This was needed for when we weren't explicitly detaching on clone
                # cmd_for_branch = self.__git + ' reflog HEAD -n 1'
                # reflog_output = shellcmd.run(shlex.split(cmd_for_branch), output=True)
                # name = reflog_output.split()[-1].strip()
                name = output.split()[-1].strip()
                tYpe = "b"
        elif output.startswith("HEAD"):  # Assume hash
            cmd = self.__git + " rev-parse HEAD"
            hash_out = shellcmd.run(shlex.split(cmd), output=True)
            detached = True
            name = hash_out.rstrip()
            tYpe = "h"
        elif output.startswith("grafted"):
            cmd = self.__git + " describe --always"
            hash_out = shellcmd.run(shlex.split(cmd), output=True)
            detached = True
            name = hash_out.rstrip()
            tYpe = "h"
        return (name, tYpe, detached)



from importlib import import_module

from mepo.cmdline.parser import MepoArgParser
from mepo.utilities import mepoconfig


def main():
    args = MepoArgParser().parse()
    mepo_cmd = mepoconfig.get_alias_command(args.mepo_cmd)

    # Load the module containing the "run" method of specified command
    cmd_module = import_module(f"mepo.command.{mepo_cmd}")

    # Execute "run" method of specified command
    cmd_module.run(args)


if __name__ == "__main__":
    main()



from collections import namedtuple

MepoVersion = namedtuple("MepoVersion", ["name", "type", "detached"])


def version_to_string(version, git=None):
    version_name = version[0]
    version_type = version[1]
    version_detached = version[2]

    if version_detached:  # detached head
        # We remove the "origin/" from the internal detached branch name
        # for clarity in mepo status output
        version_name = version_name.replace("origin/", "")
        if version_type == "b" and git:
            cur_hash = git.rev_parse(short=True).strip()
            s = f"({version_type}) {version_name} (DH, {cur_hash})"
        else:
            s = f"({version_type}) {version_name} (DH)"
    else:
        s = f"({version_type}) {version_name}"
    return s


def sanitize_version_string(orig, curr, git):
    """
    This routine tries to figure out if two tags are the same.

    The issue is that git sometimes returns the "wrong" tag in
    the mepo sense. mepo might have checked out tag v1.0.0 but
    if that commit is also tagged with foo, then sometimes mepo
    will say that things have changed because it thinks it's on
    foo.
    """

    # The trick below only works on tags and hashes (I think), so
    # if not a tag or hash, just do nothing for now
    is_tag = "(t)"
    is_hash = "(h)"

    # For status, we pass in space-delimited strings that are:
    #  'type version dh'
    # So let's split into lists and pull the type
    # But for save, we are passing in one single string
    orig_list = orig.split()
    if len(orig_list) > 1:
        # Pull out the type
        orig_type = orig_list[0]
        # Pull out the version string...
        orig_ver = orig_list[1]
    else:
        # Assume tag?
        orig_type = is_tag
        # version is the only element
        orig_ver = orig_list[0]

    curr_list = curr.split()
    if len(curr_list) > 1:
        # Pull out the type
        curr_type = curr_list[0]
        # Pull out the version string...
        curr_ver = curr_list[1]
    else:
        # Assume tag?
        curr_type = is_tag
        # version is the only element
        curr_ver = curr_list[0]

    orig_type_is_tag_or_hash = orig_type == is_tag or orig_type == is_hash
    curr_type_is_tag_or_hash = curr_type == is_tag or curr_type == is_hash

    # Now if a type or hash...
    if orig_type_is_tag_or_hash and curr_type_is_tag_or_hash:

        # Use rev-list to get the hash of the tag...
        orig_rev = git.rev_list(orig_ver)
        curr_rev = git.rev_list(curr_ver)

        # If they are identical...
        if orig_rev == curr_rev:

            # Replace the curr version with the original to make
            # mepo happy...
            curr_list[curr_list.index(curr_ver)] = orig_ver
            if orig_type == is_hash:
                curr_list[curr_list.index(curr_type)] = orig_type

            # And then remake the curr string
            curr = " ".join(curr_list)

    # And return curr
    return curr



from .branch_list import run as branch_list_run
from .branch_create import run as branch_create_run
from .branch_delete import run as branch_delete_run


def run(args):
    d = {
        "list": branch_list_run,
        "create": branch_create_run,
        "delete": branch_delete_run,
    }
    d[args.mepo_branch_cmd](args)



import os
import sys
import json
import glob
import stat
import pickle

from .registry import Registry
from .component import MepoComponent
from .utilities import colors
from .utilities.exceptions import StateDoesNotExistError
from .utilities.exceptions import StateAlreadyInitializedError
from .utilities.chdir import chdir as mepo_chdir


class MepoState(object):

    __state_dir_name = ".mepo"
    __state_fileptr_name = "state.json"
    __state_fileptr_name_old = "state.pkl"

    @staticmethod
    def get_parent_dirs():
        mypath = os.getcwd()
        parentdirs = [mypath]
        while mypath != "/":
            mypath = os.path.dirname(mypath)
            parentdirs.append(mypath)
        return parentdirs

    @classmethod
    def get_dir(cls):
        """Return location of mepo state dir"""
        for mydir in cls.get_parent_dirs():
            state_dir = os.path.join(mydir, cls.__state_dir_name)
            if os.path.exists(state_dir):
                return state_dir
        raise OSError("mepo state dir [.mepo] does not exist")

    @classmethod
    def get_root_dir(cls):
        """Return fixture (root) directory that contains mepo state dir"""
        return os.path.dirname(cls.get_dir())

    @classmethod
    def get_file(cls, old_style=False):
        """Return location of mepo state file"""
        if old_style:
            fileptr_name = cls.__state_fileptr_name_old
        else:
            fileptr_name = cls.__state_fileptr_name
        state_file = os.path.join(cls.get_dir(), fileptr_name)
        if os.path.exists(state_file):
            return state_file
        else:
            raise OSError(f"mepo state file [{state_file}] does not exist")

    @classmethod
    def state_exists(cls, old_style=False):
        try:
            cls.get_file(old_style)
            return True
        except OSError:
            return False

    @classmethod
    def initialize(cls, project_registry, directory_style):
        if cls.state_exists():
            raise StateAlreadyInitializedError("Error! mepo state already exists")
        input_components = Registry(project_registry).read_file()
        complist = list()
        for name, comp in input_components.items():
            complist.append(
                MepoComponent().registry_to_component(name, comp, directory_style)
            )
        cls.write_state(complist)

    @staticmethod
    def __mepo1_patch():
        """
        mepo1 to mepo2 includes renaming of directories
        Since pickle requires that "the class definition must be importable
        and live in the same module as when the object was stored", we need to
        patch sys.modules to be able to read mepo1 state
        """
        import mepo

        sys.modules["state"] = mepo.state
        sys.modules["state.component"] = mepo.component
        sys.modules["utilities"] = mepo.utilities

    @staticmethod
    def mepo1_patch_undo():
        """
        Undo changes made my __mepo1_patch(). Called during <mepo update-state>
        """
        entries_to_remove = ["state", "state.component", "utilities"]
        for key in entries_to_remove:
            sys.modules.pop(key, None)

    @classmethod
    def read_state(cls):
        if cls.state_exists():
            with open(cls.get_file(), "r") as fin:
                allcomps_s = json.load(fin)
            # List of dicts -> state (list of MepoComponent objects)
            allcomps = []
            for comp_s in allcomps_s:
                comp = MepoComponent().deserialize(comp_s)
                # Relative path to absolute
                comp.local = os.path.join(cls.get_root_dir(), comp.local)
                allcomps.append(comp)
            return allcomps
        elif cls.state_exists(old_style=True):
            print(
                colors.YELLOW
                + "Detected mepo1 style state\n"
                + "Run <mepo update-state> to permanently convert to mepo2 style"
                + colors.RESET
            )
            cls.__mepo1_patch()
            with open(cls.get_file(old_style=True), "rb") as fin:
                allcomps = pickle.load(fin)
            for comp in allcomps:
                comp.local = os.path.join(cls.get_root_dir(), comp.local)
        else:
            raise StateDoesNotExistError("Error! mepo state does not exist")
        return allcomps

    @classmethod
    def __get_new_state_file(cls):
        """Return full path to the new state file to write to"""
        if cls.state_exists():
            state_dir = cls.get_dir()
            pattern = os.path.join(cls.get_dir(), "state.*.json")
            states = [os.path.basename(x) for x in glob.glob(pattern)]
            new_state_id = max([int(x.split(".")[1]) for x in states]) + 1
            state_filename = "state." + str(new_state_id) + ".json"
        elif cls.state_exists(old_style=True):
            state_dir = cls.get_dir()
            pattern = os.path.join(cls.get_dir(), "state.*.pkl")
            states = [os.path.basename(x) for x in glob.glob(pattern)]
            new_state_id = max([int(x.split(".")[1]) for x in states])
            state_filename = "state." + str(new_state_id) + ".json"
        else:
            state_dir = os.path.join(os.getcwd(), cls.__state_dir_name)
            os.makedirs(state_dir, exist_ok=True)
            state_filename = "state.0.json"
        return os.path.join(state_dir, state_filename)

    @classmethod
    def write_state(cls, allcomps):
        new_state_file = cls.__get_new_state_file()
        allcomps_s = []
        for comp in allcomps:
            # Save relative path (to fixture dir) to state
            comp.local = os.path.relpath(comp.local, start=cls.get_root_dir())
            allcomps_s.append(comp.serialize())
        with open(new_state_file, "w") as fout:
            json.dump(allcomps_s, fout)
        # Make the state file read-only
        os.chmod(new_state_file, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
        # Update symlink
        state_dir = os.path.dirname(new_state_file)
        state_fileptr_fullpath = os.path.join(state_dir, cls.__state_fileptr_name)
        if os.path.isfile(state_fileptr_fullpath):
            os.remove(state_fileptr_fullpath)
        with mepo_chdir(state_dir):
            new_state_filename = os.path.basename(new_state_file)
            os.symlink(new_state_filename, cls.__state_fileptr_name)



from ..state import MepoState
from ..utilities import verify
from ..git import GitRepository


def run(args):
    allcomps = MepoState.read_state()
    comps2list = _get_comps_to_list(args.comp_name, allcomps)
    max_namelen = len(max([x.name for x in comps2list], key=len))
    FMT = "{:<%s.%ss} | {:<s}" % (max_namelen, max_namelen)
    for comp in comps2list:
        git = GitRepository(comp.remote, comp.local)
        output = git.list_tags().rstrip().split("\n")
        print(FMT.format(comp.name, output[0]))
        for line in output[1:]:
            print(FMT.format("", line))


def _get_comps_to_list(specified_comps, allcomps):
    comps_to_list = allcomps
    if specified_comps:
        verify.valid_components(specified_comps, allcomps)
        comps_to_list = [x for x in allcomps if x.name in specified_comps]
    return comps_to_list
